/*
 * Copyright 2016-2017 Golden Code Development Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"),  to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

package com.goldencode.hotel;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.hotel.dmo._temp.*;
import com.goldencode.hotel.dmo.hotel.*;
import com.goldencode.p2j.persist.lock.*;

import static com.goldencode.p2j.util.BlockManager.*;
import static com.goldencode.p2j.util.CompareOps.*;
import static com.goldencode.p2j.util.MathOps.*;
import static com.goldencode.p2j.util.logical.*;
import static com.goldencode.p2j.util.TextOps.*;
import static com.goldencode.p2j.util.character.*;
import static com.goldencode.p2j.persist.RecordBuffer.batch;
import static com.goldencode.p2j.util.ErrorManager.*;

/**
 * Business logic (converted to Java from the 4GL source code
 * in ehotel.p).
 */
public class Ehotel
{
   public static final String FMT_STR_1 = "z9.99";

   /* temp-table of aggregated data for room types over a period */   
   TtRoomTypeAgg_1_1.Buf ttRoomTypeAgg = TemporaryBuffer.define(TtRoomTypeAgg_1_1.Buf.class, "ttRoomTypeAgg", "tt-room-type-agg", false);

   /*
    * Implementation of ABL procedures which comprise an API exposed to JavaScript. The naming
    * convention for procedures which implement up-calls from JavaScript into converted ABL is to
    * use the prefix "inMsg_".
    *
    * Each such procedure invokes a companion procedure whose name has the prefix "sendMsg_". This
    * is the procedure which performs the actual work of the API. When it is done with its work, the
    * "sendMsg_" procedure invokes a JavaScript callback function based on a unique message ID
    * provided in the originating up-call. Data gathered by the API is passed as a JSON object to
    * the callback function via P2J-REMOTE-CALL.
    *
    * This file contains two examples of APIs which perform database access and "return" row data
    * within JSON objects. It also contains a procedure which is driven by a JavaScript navigation
    * bar to select the embedded ADM2 window to be displayed when the user clicks on a menu item in
    * the navigation bar.
    *
    * TODO: push the up-call/JSON scaffolding code up into the FWD runtime, so that only the core
    * functionality is implemented in the ABL here.
    */   
   /* temp-table to count rooms utilized (reserved or occupied) by date */   
   TtRoomsUsed_1_1.Buf ttRoomsUsed = TemporaryBuffer.define(TtRoomsUsed_1_1.Buf.class, "ttRoomsUsed", "tt-rooms-used", false);

   Rate.Buf rate = RecordBuffer.define(Rate.Buf.class, "hotel", "rate", "rate");

   RoomType.Buf roomType = RecordBuffer.define(RoomType.Buf.class, "hotel", "roomType", "room-type");

   Stay.Buf stay = RecordBuffer.define(Stay.Buf.class, "hotel", "stay", "stay");

   Reservation.Buf reservation = RecordBuffer.define(Reservation.Buf.class, "hotel", "reservation", "reservation");

   Room.Buf room = RecordBuffer.define(Room.Buf.class, "hotel", "room", "room");

   /**
    * External procedure (converted to Java from the 4GL source code
    * in ehotel.p).
    */
   public void execute()
   {
      externalProcedure(Ehotel.this, new Block((Body) () -> 
      {
         RecordBuffer.openScope(ttRoomsUsed, ttRoomTypeAgg, room, reservation, stay, roomType, rate);
      }));
   }

   /**
    * Up-call from JavaScript to gather hotel room utilization data for a range of dates.<br>
    * <p>
    * Parameters:<br>
    * <p>
    *    messageId (input)    - Unique message ID for the up-call.<br>
    *    start-date (input)   - Start of date range.<br>
    *    end-date (input)     - End of date range.
    */
   public void inMsgUtilizationByDate(final integer _messageId, final date _startDate, final date _endDate)
   {
      integer messageId = UndoableFactory.initInput(_messageId);
      date startDate = UndoableFactory.initInput(_startDate);
      date endDate = UndoableFactory.initInput(_endDate);
      
      internalProcedure(new Block((Body) () -> 
      {
         ControlFlowOps.invokeWithMode("sendMsg_utilizationByDate", "III", messageId, startDate, endDate);
      }));
   }

   /**
    * Collect hotel room utilization data for a range of dates and supply that data back to<br>
    * JavaScript code as a JSON message, via a remote call.<br>
    * <p>
    * Parameters:<br>
    * <p>
    *    messageId (input)    - Unique message ID for the up-call.<br>
    *    start-date (input)   - Start of date range.<br>
    *    end-date (input)     - End of date range.
    */
   public void sendMsgUtilizationByDate(final integer _messageId, final date _startDate, final date _endDate)
   {
      integer messageId = UndoableFactory.initInput(_messageId);
      date startDate = UndoableFactory.initInput(_startDate);
      date endDate = UndoableFactory.initInput(_endDate);
      date nextDate = TypeFactory.date();
      date d1 = TypeFactory.date();
      date d2 = TypeFactory.date();
      decimal useCount = UndoableFactory.decimal();
      decimal totalRooms = UndoableFactory.decimal(new decimal(0));
      decimal full = UndoableFactory.decimal();
      character jsonMsg = UndoableFactory.character();
      character rowData = UndoableFactory.character();
      logical comma = UndoableFactory.logical(false);
      
      internalProcedure(new Block((Init) () -> 
      {
         full.setPrecision(4);
      }, 
      (Body) () -> 
      {
         AdaptiveQuery query0 = new AdaptiveQuery();
         
         forEach("loopLabel0", new Block((Init) () -> 
         {
            query0.initialize(room, ((String) null), null, "room.roomNum asc", LockType.NONE);
         }, 
         (Body) () -> 
         {
            query0.next();
            totalRooms.assign(plus(totalRooms, 1));
         }));
         
         AdaptiveQuery query1 = new AdaptiveQuery();
         
         forEach("loopLabel1", new Block((Init) () -> 
         {
            query1.initialize(reservation, "reservation.startDate <= ? and reservation.endDate > ?", null, "reservation.roomNum asc, reservation.startDate asc, reservation.endDate asc", new Object[]
            {
               endDate,
               startDate
            }, LockType.NONE);
         }, 
         (Body) () -> 
         {
            /* collect reservations; include check-in date, exclude check-out date */
            query1.next();
            d1.assign((_isGreaterThan(reservation.getStartDate(), startDate) ? reservation.getStartDate() : startDate));
            d2.assign((_isLessThanOrEqual(reservation.getEndDate(), endDate) ? date.minusDays(reservation.getEndDate(), 1) : endDate));
            
            loopLabel2:
            for (nextDate.assign(d1); _isLessThanOrEqual(nextDate, d2); nextDate.increment())
            {
               if (_not(new FindQuery(ttRoomsUsed, "ttRoomsUsed.useDate = ? and ttRoomsUsed.roomNum = ?", null, "ttRoomsUsed.useDate asc, ttRoomsUsed.roomNum asc", new Object[]
               {
                  nextDate,
                  (P2JQuery.Parameter) () -> reservation.getRoomNum()
               }, LockType.NONE).hasOne()))
               {
                  ttRoomsUsed.create();
                  batch(() -> 
                  {
                     ttRoomsUsed.setUseDate(nextDate);
                     ttRoomsUsed.setRoomNum(reservation.getRoomNum());
                  });
               }
            }
         }));
         
         AdaptiveQuery query2 = new AdaptiveQuery();
         
         forEach("loopLabel3", new Block((Init) () -> 
         {
            query2.initialize(stay, "stay.startDate <= ? and stay.endDate > ?", null, "stay.stayId asc", new Object[]
            {
               endDate,
               startDate
            }, LockType.NONE);
         }, 
         (Body) () -> 
         {
            /* collect stays; include check-in date, exclude check-out date */
            query2.next();
            d1.assign((_isGreaterThan(stay.getStartDate(), startDate) ? stay.getStartDate() : startDate));
            d2.assign((_isLessThanOrEqual(stay.getEndDate(), endDate) ? date.minusDays(stay.getEndDate(), 1) : endDate));
            
            loopLabel4:
            for (nextDate.assign(d1); _isLessThanOrEqual(nextDate, d2); nextDate.increment())
            {
               if (_not(new FindQuery(ttRoomsUsed, "ttRoomsUsed.useDate = ? and ttRoomsUsed.roomNum = ?", null, "ttRoomsUsed.useDate asc, ttRoomsUsed.roomNum asc", new Object[]
               {
                  nextDate,
                  (P2JQuery.Parameter) () -> stay.getRoomNum()
               }, LockType.NONE).hasOne()))
               {
                  ttRoomsUsed.create();
                  batch(() -> 
                  {
                     ttRoomsUsed.setUseDate(nextDate);
                     ttRoomsUsed.setRoomNum(stay.getRoomNum());
                  });
               }
            }
         }));
         
         PresortQuery query3 = new PresortQuery();
         FieldReference byExpr0 = new FieldReference(ttRoomsUsed, "useDate");
         
         forEach("loopLabel5", new Block((Init) () -> 
         {
            query3.initialize(ttRoomsUsed, ((String) null), null, "ttRoomsUsed.useDate asc, ttRoomsUsed.roomNum asc");
            query3.enableBreakGroups();
            query3.setNonScrolling();
            query3.addSortCriterion(byExpr0);
         }, 
         (Body) () -> 
         {
            /* count up used rooms by date */
            query3.next();
            
            if ((query3.isFirstOfGroup(byExpr0)).booleanValue())
            {
               useCount.assign(0);
            }
            
            useCount.assign(plus(useCount, 1));
            
            if ((query3.isLastOfGroup(byExpr0)).booleanValue())
            {
               /* treat overbookings as fully reserved */
               if (_isGreaterThan(useCount, totalRooms))
               {
                  useCount.assign(totalRooms);
               }
               
               rowData.assign(concat(rowData, ((comma).booleanValue() ? new character(",\n") : new character(""))));
               full.assign(divide(useCount, totalRooms));
               rowData.assign(concat(rowData, "{"));
               rowData.assign(concat(rowData, "\"date\":\"", valueOf(ttRoomsUsed.getUseDate(), "99/99/9999"), "\","));
               rowData.assign(concat(rowData, "\"score\":", valueOf(full, "z9.9999")));
               rowData.assign(concat(rowData, "}"));
               comma.assign(true);
            }
         }));
         
         silent(() -> ttRoomsUsed.deleteAll());
         jsonMsg.assign("{\n");
         jsonMsg.assign(concat(jsonMsg, " \"api\":\"utilizationByDateResponse\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"messageId\":\"", valueOf(messageId), "\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"rows\": [", rowData, "],\n"));
         jsonMsg.assign(concat(jsonMsg, " \"domain\": [ 0.0, 1.0 ],\n"));
         jsonMsg.assign(concat(jsonMsg, " \"start\":\"", valueOf(startDate, "99/99/9999"), "\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"end\":\"", valueOf(endDate, "99/99/9999"), "\"\n"));
         jsonMsg.assign(concat(jsonMsg, "}"));
         ControlFlowOps.remoteCall(jsonMsg);
      }));
   }

   /**
    * Up-call from JavaScript to gather hotel room utilization data, grouped by room type, for a<br>
    * range of dates.<br>
    * <p>
    * Parameters:<br>
    * <p>
    *    messageId (input)    - Unique message ID for the up-call.<br>
    *    start-date (input)   - Start of date range.<br>
    *    end-date (input)     - End of date range.
    */
   public void inMsgUtilizationByType(final integer _messageId, final date _startDate, final date _endDate)
   {
      integer messageId = UndoableFactory.initInput(_messageId);
      date startDate = UndoableFactory.initInput(_startDate);
      date endDate = UndoableFactory.initInput(_endDate);
      
      internalProcedure(new Block((Body) () -> 
      {
         ControlFlowOps.invokeWithMode("sendMsg_utilizationByType", "III", messageId, startDate, endDate);
      }));
   }

   /**
    * Collect hotel room utilization data, grouped by room type, for a range of dates. Supply that<br>
    * data back to JavaScript code as a JSON message, via a remote call.<br>
    * <p>
    * Parameters:<br>
    * <p>
    *    messageId (input)    - Unique message ID for the up-call.<br>
    *    start-date (input)   - Start of date range.<br>
    *    end-date (input)     - End of date range.
    */
   public void sendMsgUtilizationByType(final integer _messageId, final date _startDate, final date _endDate)
   {
      integer messageId = UndoableFactory.initInput(_messageId);
      date startDate = UndoableFactory.initInput(_startDate);
      date endDate = UndoableFactory.initInput(_endDate);
      character jsonMsg = TypeFactory.character();
      character msg = TypeFactory.character("");
      decimal weight = TypeFactory.decimal();
      decimal score = TypeFactory.decimal();
      integer totalRoomCount = TypeFactory.integer((long) 0);
      decimal actualCount = TypeFactory.decimal();
      decimal revenue = TypeFactory.decimal();
      decimal totalPotRev = TypeFactory.decimal(new decimal(0));
      decimal actualRevenue = TypeFactory.decimal();
      date d1 = TypeFactory.date();
      date d2 = TypeFactory.date();
      integer dayCount = TypeFactory.integer();
      date nextDate = TypeFactory.date();
      logical comma = TypeFactory.logical(false);
      logical rateWeighted = TypeFactory.logical(true);
      decimal r1 = TypeFactory.decimal();
      final CountAccumulator accumCount0 = new CountAccumulator();
      
      internalProcedure(new Block((Init) () -> 
      {
         totalPotRev.setPrecision(2);
         actualRevenue.setPrecision(2);
         r1.setPrecision(2);
      }, 
      (Body) () -> 
      {
         PresortQuery query4 = new PresortQuery();
         FieldReference byExpr1 = new FieldReference(room, "roomType");
         
         forEach("loopLabel6", new Block((Init) () -> 
         {
            query4.initialize();
            query4.enableBreakGroups();
            query4.setNonScrolling();
            query4.addComponent(roomType, ((String) null), LockType.NONE);
            query4.addComponent(room, "room.roomType = roomType.roomType and room.active = true", LockType.NONE);
            query4.addSortCriterion(byExpr1);
            query4.addAccumulator(accumCount0);
            accumCount0.setSubOnly();
            accumCount0.addBreakGroup(byExpr1);
            accumCount0.reset();
         }, 
         (Body) () -> 
         {
            /* count the number of active rooms of each type */
            query4.next();
            accumCount0.accumulate();
            
            if ((query4.isLastOfGroup(byExpr1)).booleanValue())
            {
               ttRoomTypeAgg.create();
               batch(() -> 
               {
                  ttRoomTypeAgg.setRoomType(room.getRoomType());
                  ttRoomTypeAgg.setCount(accumCount0.getResult(byExpr1));
                  ttRoomTypeAgg.setDescription(roomType.getDescription());
               });
               totalRoomCount.assign(plus(totalRoomCount, ttRoomTypeAgg.getCount()));
            }
         }));
         
         /* get total count of rooms that could potentially be rented for the period */
         dayCount.assign(plus(date.differenceNum(endDate, startDate), 1));
         totalRoomCount.assign(multiply(totalRoomCount, dayCount));
         
         CompoundQuery query5 = new CompoundQuery();
         
         forEach("loopLabel7", new Block((Init) () -> 
         {
            query5.initialize(false, false);
            query5.addComponent(new AdaptiveQuery().initialize(ttRoomTypeAgg, ((String) null), null, "ttRoomTypeAgg.roomType asc"));
            query5.addComponent(new AdaptiveQuery().initialize(rate, "rate.roomType = ? and rate.startDate <= ? and rate.endDate >= ?", null, "rate.roomType asc, rate.startDate asc, rate.endDate asc", new Object[]
            {
               new FieldReference(ttRoomTypeAgg, "roomType"),
               endDate,
               startDate
            }, LockType.NONE));
         }, 
         (Body) () -> 
         {
            /* calculate potential revenue for the period by room type using the associated room rates */
            /* note: this algorithm assumes rates do not overlap! */
            query5.iterate();
            d1.assign((_isGreaterThan(rate.getStartDate(), startDate) ? rate.getStartDate() : startDate));
            d2.assign((_isLessThanOrEqual(rate.getEndDate(), endDate) ? rate.getEndDate() : endDate));
            dayCount.assign(plus(date.differenceNum(d2, d1), 1));
            revenue.assign(multiply(multiply(dayCount, rate.getRate()), ttRoomTypeAgg.getCount()));
            ttRoomTypeAgg.setPotentialRevenue(plus(ttRoomTypeAgg.getPotentialRevenue(), revenue));
            totalPotRev.assign(plus(totalPotRev, revenue));
         }));
         
         CompoundQuery query6 = new CompoundQuery();
         
         forEach("loopLabel8", new Block((Init) () -> 
         {
            query6.initialize(false, false);
            query6.addComponent(new AdaptiveQuery().initialize(room, "room.active = true", null, "room.roomNum asc", LockType.NONE));
            query6.addComponent(new AdaptiveQuery().initialize(reservation, "reservation.roomNum = ? and reservation.startDate <= ? and reservation.endDate > ?", null, "reservation.roomNum asc, reservation.startDate asc, reservation.endDate asc", new Object[]
            {
               new FieldReference(room, "roomNum"),
               endDate,
               startDate
            }, LockType.NONE));
         }, 
         (Body) () -> 
         {
            /* collect daily room utilization data for each reservation for the period */
            query6.iterate();
            d1.assign((_isGreaterThan(reservation.getStartDate(), startDate) ? reservation.getStartDate() : startDate));
            d2.assign((_isLessThanOrEqual(reservation.getEndDate(), endDate) ? date.minusDays(reservation.getEndDate(), 1) : endDate));
            
            loopLabel9:
            for (nextDate.assign(d1); _isLessThanOrEqual(nextDate, d2); nextDate.increment())
            {
               r1.assign(0);
               
               if ((rateWeighted).booleanValue())
               {
                  silent(() -> new FindQuery(rate, "rate.roomType = ? and rate.startDate <= ? and rate.endDate >= ?", null, "rate.roomType asc, rate.startDate asc, rate.endDate asc", new Object[]
                  {
                     (P2JQuery.Parameter) () -> room.getRoomType(),
                     nextDate,
                     nextDate
                  }, LockType.NONE).first());
                  
                  if (rate._available())
                  {
                     r1.assign(rate.getRate());
                  }
                  else
                  {
                     rateWeighted.assign(false);
                  }
               }
               ttRoomsUsed.create();
               batch(() -> 
               {
                  ttRoomsUsed.setUseDate(nextDate);
                  ttRoomsUsed.setRoomNum(reservation.getRoomNum());
                  ttRoomsUsed.setRoomType(room.getRoomType());
                  ttRoomsUsed.setRate(r1);
               });
            }
         }));
         
         CompoundQuery query7 = new CompoundQuery();
         
         forEach("loopLabel10", new Block((Init) () -> 
         {
            query7.initialize(false, false);
            query7.addComponent(new AdaptiveQuery().initialize(room, "room.active = true", null, "room.roomNum asc", LockType.NONE));
            query7.addComponent(new AdaptiveQuery().initialize(stay, "stay.roomNum = ? and stay.startDate <= ? and stay.endDate > ?", null, "stay.stayId asc", new Object[]
            {
               new FieldReference(room, "roomNum"),
               endDate,
               startDate
            }, LockType.NONE));
         }, 
         (Body) () -> 
         {
            /* collect daily room utilization data for each stay for the period */
            query7.iterate();
            d1.assign((_isGreaterThan(stay.getStartDate(), startDate) ? stay.getStartDate() : startDate));
            d2.assign((_isLessThanOrEqual(stay.getEndDate(), endDate) ? date.minusDays(stay.getEndDate(), 1) : endDate));
            
            loopLabel11:
            for (nextDate.assign(d1); _isLessThanOrEqual(nextDate, d2); nextDate.increment())
            {
               r1.assign(0);
               
               if ((rateWeighted).booleanValue())
               {
                  silent(() -> new FindQuery(rate, "rate.roomType = ? and rate.startDate <= ? and rate.endDate >= ?", null, "rate.roomType asc, rate.startDate asc, rate.endDate asc", new Object[]
                  {
                     (P2JQuery.Parameter) () -> room.getRoomType(),
                     nextDate,
                     nextDate
                  }, LockType.NONE).first());
                  
                  if (rate._available())
                  {
                     r1.assign(rate.getRate());
                  }
                  else
                  {
                     rateWeighted.assign(false);
                  }
               }
               ttRoomsUsed.create();
               batch(() -> 
               {
                  ttRoomsUsed.setUseDate(nextDate);
                  ttRoomsUsed.setRoomNum(stay.getRoomNum());
                  ttRoomsUsed.setRoomType(room.getRoomType());
                  ttRoomsUsed.setRate(r1);
               });
            }
         }));
         
         AdaptiveQuery query8 = new AdaptiveQuery();
         
         forEach("loopLabel12", new Block((Init) () -> 
         {
            query8.initialize(ttRoomTypeAgg, ((String) null), null, "ttRoomTypeAgg.roomType asc");
         }, 
         (Body) () -> 
         {
            /* walk the collected, aggregate data and compose summary data for each room type in JSON */
            query8.next();
            actualRevenue.assign(0);
            actualCount.assign(0);
            
            AdaptiveQuery query9 = new AdaptiveQuery();
            
            forEach("loopLabel13", new Block((Init) () -> 
            {
               query9.initialize(ttRoomsUsed, "ttRoomsUsed.roomType = ?", null, "ttRoomsUsed.roomType asc", new Object[]
               {
                  (P2JQuery.Parameter) () -> ttRoomTypeAgg.getRoomType()
               });
            }, 
            (Body) () -> 
            {
               query9.next();
               actualRevenue.assign(plus(actualRevenue, ttRoomsUsed.getRate()));
               actualCount.assign(plus(actualCount, 1));
            }));
            
            if ((rateWeighted).booleanValue())
            {
               weight.assign(multiply(divide(ttRoomTypeAgg.getPotentialRevenue(), totalPotRev), 100));
               score.assign(multiply(divide(actualRevenue, ttRoomTypeAgg.getPotentialRevenue()), 100));
            }
            else
            {
               weight.assign(multiply(divide(ttRoomTypeAgg.getCount(), totalRoomCount), 100));
               score.assign(multiply(divide(actualCount, multiply(ttRoomTypeAgg.getCount(), dayCount)), 100));
            }
            
            msg.assign(concat(msg, ((comma).booleanValue() ? new character(",\n") : new character(""))));
            msg.assign(concat(msg, "{"));
            msg.assign(concat(msg, "\"id\":\"", ttRoomTypeAgg.getRoomType(), "\","));
            msg.assign(concat(msg, "\"score\":\"", valueOf(score, "zz9.99"), "\","));
            msg.assign(concat(msg, "\"weight\":\"", valueOf(weight, "zz9.99"), "\","));
            msg.assign(concat(msg, "\"actualRevenue\":\"", actualRevenue, "\","));
            msg.assign(concat(msg, "\"potentialRevenue\":\"", ttRoomTypeAgg.getPotentialRevenue(), "\","));
            msg.assign(concat(msg, "\"roomsOccupied\":\"", valueOf(actualCount, "zzzzz9"), "\","));
            msg.assign(concat(msg, "\"totalRooms\":\"", valueOf(multiply(ttRoomTypeAgg.getCount(), dayCount), "zzzzz9"), "\","));
            msg.assign(concat(msg, "\"label\":\"", ttRoomTypeAgg.getDescription(), "\""));
            msg.assign(concat(msg, "}"));
            comma.assign(true);
         }));
         
         jsonMsg.assign("{\n");
         jsonMsg.assign(concat(jsonMsg, " \"api\":\"utilizationByTypeResponse\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"messageId\":\"", valueOf(messageId), "\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"rows\": [\n", msg, "],\n"));
         jsonMsg.assign(concat(jsonMsg, " \"rateWeighted\":\"", ((rateWeighted).booleanValue() ? new character("true") : new character("false")), "\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"start\":\"", valueOf(startDate, "99/99/9999"), "\",\n"));
         jsonMsg.assign(concat(jsonMsg, " \"end\":\"", valueOf(endDate, "99/99/9999"), "\"\n"));
         jsonMsg.assign(concat(jsonMsg, "}"));
         ttRoomTypeAgg.deleteAll();
         ttRoomsUsed.deleteAll();
         ControlFlowOps.remoteCall(jsonMsg);
      }));
   }

   public character fwdGetAdmversion()
   {
      return function("fwdGetADMVersion", character.class, new Block((Body) () -> 
      {
         returnNormal("ADM2.2");
      }));
   }
}
