/*
** Module   : FwdEmbeddedDriver.java
** Abstract : FWD driver for running ADM windows in embedded mode.
**
** Copyright (c) 2017-2018, Golden Code Development Corporation.
**
** _#_ _I_ __Date__ ________________________________Description___________________________________
** 001 CA  20170825 Created initial version.
** 002 CA  20170928 Use destroyObject/destroy to clear a window resources and other improvements.
** 003 CA  20171004 Added APIs to expose legacy menus to the embedded driver (both definition and 
**                  event support).
** 004 CA  20171023 ADM version is kept on a per-window basis.
** 005 CA  20171109 Some improvements when initializing the embedded window's page.
** 006 CA  20171111 Allow default WINDOW:BGCOLOR-RGB to be set. Center ABL windows horizontally.
**                  Don't move the frame's column position, just row.
** 007 HC  20171206 Added fwd-user-logout shared variable that can be used as a logout condition.
** 008 CA  20180313 Added security support.  Added module and toolbar support.
** 009 CA  20180727 Fixes to better position ADM frames in embedded mode.
**                  Added notification that the business logic has switched the ADM page.
**                  Allow the JS side to hide the window status and/or message area for the 
**                  4GL windows handled via this driver.
** 010 CA  20180801 Allow multiple ADM windows to be shown at the same time, in embedded mode.
** 011 CA  20180814 Added fwd-embedded-mode var.
*/

package com.goldencode.hotel;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.persist.*;
import com.goldencode.hotel.dmo._temp.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.p2j.persist.lock.*;

import static com.goldencode.p2j.util.BlockManager.*;
import static com.goldencode.p2j.util.CompareOps.*;
import static com.goldencode.p2j.util.TextOps.*;
import static com.goldencode.p2j.util.MathOps.*;
import static com.goldencode.p2j.util.logical.*;
import static com.goldencode.p2j.util.ProcedureManager.*;
import static com.goldencode.p2j.util.ArrayAssigner.*;
import static com.goldencode.p2j.util.character.*;
import static com.goldencode.p2j.ui.LogicalTerminal.*;
import static com.goldencode.p2j.util.ErrorManager.*;

/**
 * Business logic (converted to Java from the 4GL source code
 * in fwd-embedded-driver.p).
 */
public class FwdEmbeddedDriver
{
   PageToolbar_1_1.Buf pageToolbar = TemporaryBuffer.define(PageToolbar_1_1.Buf.class, "pageToolbar", "page-toolbar", false);

   FwdEmbeddedWindow_1_1.Buf fwdEmbeddedWindow = SharedVariableManager.addTempTable(ScopeLevel.GLOBAL, "fwdEmbeddedWindow", FwdEmbeddedWindow_1_1.Buf.class, "fwdEmbeddedWindow", "fwd-embedded-window");

   /*
   ** This program is free software: you can redistribute it and/or modify
   ** it under the terms of the GNU Affero General Public License as
   ** published by the Free Software Foundation, either version 3 of the
   ** License, or (at your option) any later version.
   **
   ** This program is distributed in the hope that it will be useful,
   ** but WITHOUT ANY WARRANTY; without even the implied warranty of
   ** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   ** GNU Affero General Public License for more details.
   **
   ** You may find a copy of the GNU Affero GPL version 3 at the following
   ** location: https://www.gnu.org/licenses/agpl-3.0.en.html
   ** 
   ** Additional terms under GNU Affero GPL version 3 section 7:
   ** 
   **   Under Section 7 of the GNU Affero GPL version 3, the following additional
   **   terms apply to the works covered under the License.  These additional terms
   **   are non-permissive additional terms allowed under Section 7 of the GNU
   **   Affero GPL version 3 and may not be removed by you.
   ** 
   **   0. Attribution Requirement.
   ** 
   **     You must preserve all legal notices or author attributions in the covered
   **     work or Appropriate Legal Notices displayed by works containing the covered
   **     work.  You may not remove from the covered work any author or developer
   **     credit already included within the covered work.
   ** 
   **   1. No License To Use Trademarks.
   ** 
   **     This license does not grant any license or rights to use the trademarks
   **     Golden Code, FWD, any Golden Code or FWD logo, or any other trademarks
   **     of Golden Code Development Corporation. You are not authorized to use the
   **     name Golden Code, FWD, or the names of any author or contributor, for
   **     publicity purposes without written authorization.
   ** 
   **   2. No Misrepresentation of Affiliation.
   ** 
   **     You may not represent yourself as Golden Code Development Corporation or FWD.
   ** 
   **     You may not represent yourself for publicity purposes as associated with
   **     Golden Code Development Corporation, FWD, or any author or contributor to
   **     the covered work, without written authorization.
   ** 
   **   3. No Misrepresentation of Source or Origin.
   ** 
   **     You may not represent the covered work as solely your work.  All modified
   **     versions of the covered work must be marked in a reasonable way to make it
   **     clear that the modified work is not originating from Golden Code Development
   **     Corporation or FWD.  All modified versions must contain the notices of
   **     attribution required in this license.
   */   
   /* allow both explicit and implicit (via p2j.embedded.js) run of this program. */   
   handle fwdDriverHandle = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "fwd-driver-handle", new handle());

   /* set this var in the embedded mode entry point's (i.e. emain.p), to true */   
   logical fwdEmbeddedMode = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "fwd-embedded-mode", true, new logical(false));

   /* handle which refers to a procedure which (might) include custom APIs to be used for security,
      program initialization, etc */   
   handle fwdCustomApis = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "fwd-custom-apis", new handle());

   /* indicates whether user has logged out, this variable can be used as a logout condition */   
   logical fwdUserLogout = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "fwd-user-logout", new logical(false));

   integer lastPage = UndoableFactory.integer((Long) null);

   character lastWindow = UndoableFactory.character(null);

   integer lastWindowId = UndoableFactory.integer((Long) null);

   handle[] hpages = UndoableFactory.handleExtent(100);

   integer numPages = UndoableFactory.integer();

   logical fwdSelectPage = UndoableFactory.logical(false);

   integer admSelectedPage = UndoableFactory.integer((Long) null);

   /* JS-side will configure the fwd-embedded-driver if window status-area and message-area a*/   
   /* handled by JS.  For embedded mode clients, FWD will allow status-area and message-area to b*/   
   /* hidden, even if the window is realized (no error will be shown*/   
   logical jsStatusArea = UndoableFactory.logical(false);

   logical jsMessageArea = UndoableFactory.logical(false);

   /**
    * External procedure (converted to Java from the 4GL source code
    * in fwd-embedded-driver.p).
    */
   public void execute()
   {
      externalProcedure(FwdEmbeddedDriver.this, new Block((Body) () -> 
      {
         RecordBuffer.openScope(fwdEmbeddedWindow, pageToolbar);
         
         if ((fwdDriverHandle.isValid()).booleanValue())
         {
            /* if the driver was already loaded, don't run the external program code.  the APIs, even if
                  invoked via this instance, it will re-use the globally shared temp-table */
            returnNormal();
         }
         
         if (_not(thisProcedure().unwrapProcedure().isPersistent()))
         {
         }
         
         fwdDriverHandle.assign(thisProcedure());
         /* publishing will be after business logic has selected the pag*/
         /* ADM: src/adm/method/brokcntn.i - broker-select-page - will have at line 40*/
         /*     PUBLISH "fwd-adm-page-changed" (input p-caller, input p-page#*/
         /* ADM2: src/adm2/containr.p - selectpage - will have at line 459*/
         /*     PUBLISH "fwd-adm-page-changed" (input target-procedure, input iCurrentPage*/
         NamedEventManager.subscribeAnywhere(thisProcedure(), "fwd-adm-page-changed");
      }));
   }

   public void fwdAdmPageChanged(final handle _pCaller, final integer _pPageNumber)
   {
      handle pCaller = TypeFactory.initInput(_pCaller);
      integer pPageNumber = TypeFactory.initInput(_pPageNumber);
      character pfile = UndoableFactory.character();
      integer windowId = UndoableFactory.integer();
      character pages = UndoableFactory.character();
      handle hwin = UndoableFactory.handle();
      
      internalProcedure(new Block((Body) () -> 
      {
         pfile.assign(resolveProgramName(new character(pCaller.unwrap().name())));
         
         if (_isNotEqual(substring(pfile, 1, minus(length(pfile), 1)), substring(lastWindow, 1, minus(length(lastWindow), 1))))
         {
            /* some other ADM window has been selected, nothing to d*/
            if ((matchesList(pCaller.unwrapProcedure().internalEntries(), "fwdGetWindow")).booleanValue())
            {
               hwin.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetWindow", pCaller));
            }
            
            if ((hwin.isValid()).booleanValue())
            {
               windowId.assign(hwin.unwrapWidget().getHWND());
               /* notify the JS side that another page has been selected, to update the UI sta*/
               ControlFlowOps.remoteCall(concat("{ \"api\":\"admPageChanged\", \"windowId\":\"", valueOf(windowId), "\", \"page\":\"", valueOf(pPageNumber), "\" }"));
            }
            
            returnNormal();
         }
         
         /* a call from within the driver's page swit*/
         if ((fwdSelectPage).booleanValue())
         {
            /* keep the page set by A*/
            admSelectedPage.assign(pPageNumber);
            returnNormal();
         }
         
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(lastWindow)
         }).first());
         
         if (_and(fwdEmbeddedWindow.available(), () -> not(fwdEmbeddedWindow.getPhandle().isValid())))
         {
            doBlock(TransactionType.FULL, "blockLabel0", new Block((Body) () -> 
            {
               fwdEmbeddedWindow.deleteRecord();
            }));
         }
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            returnNormal();
         }
         
         /* set this n*/
         lastPage.assign(pPageNumber);
         ControlFlowOps.invokeWithMode("initEmbeddedPage", "IIIIOO", pCaller, pPageNumber, new logical(false), new unknown(), windowId, pages);
         /* notify the JS side that another page has been selected, to update the UI sta*/
         ControlFlowOps.remoteCall(concat("{ \"api\":\"admPageChanged\", \"windowId\":\"", valueOf(windowId), "\", \"page\":\"", valueOf(pPageNumber), "\" }"));
      }));
   }

   public void inMsgEmbeddedStatusArea(final logical _l)
   {
      logical l = UndoableFactory.initInput(_l);
      
      internalProcedure(new Block((Body) () -> 
      {
         jsStatusArea.assign(l);
      }));
   }

   public void inMsgEmbeddedMessageArea(final logical _l)
   {
      logical l = UndoableFactory.initInput(_l);
      
      internalProcedure(new Block((Body) () -> 
      {
         jsMessageArea.assign(l);
      }));
   }

   public void inMsgLogout()
   {
      internalProcedure(new Block((Body) () -> 
      {
         fwdUserLogout.assign(true);
         
         if (_and(fwdCustomApis.isValid(), () -> matchesList(fwdCustomApis.unwrapProcedure().internalEntries(), "fwdLogout")))
         {
            ControlFlowOps.invokeIn("fwdLogout", fwdCustomApis);
         }
         
         quit();
      }));
   }

   public character inMsgFwdGetToolBar()
   {
      handle tbhandle = UndoableFactory.handle();
      
      return function("inMsg_fwdGetToolBar", character.class, new Block((Body) () -> 
      {
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(lastWindow)
         }).first());
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            returnNormal(new character());
         }
         
         tbhandle.assign(fwdEmbeddedWindow.getToolbar());
         
         if (_not(tbhandle.isValid()))
         {
            returnNormal(new character());
         }
         
         returnNormal(fwdEmbeddedWindow.getToolbars(minus((lastPage), 1)));
      }));
   }

   public character inMsgFwdGetMenuBar()
   {
      handle mbhandle = UndoableFactory.handle();
      
      return function("inMsg_fwdGetMenuBar", character.class, new Block((Body) () -> 
      {
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(lastWindow)
         }).first());
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            returnNormal(new character());
         }
         
         mbhandle.assign(fwdEmbeddedWindow.getMenubar());
         
         if (_not(mbhandle.isValid()))
         {
            returnNormal(new character());
         }
         
         returnNormal(listMenu(mbhandle));
      }));
   }

   public character inMsgFwdGetModules()
   {
      return function("inMsg_fwdGetModules", character.class, new Block((Body) () -> 
      {
         if (_and(fwdCustomApis.isValid(), () -> matchesList(fwdCustomApis.unwrapProcedure().internalEntries(), "fwdGetModules")))
         {
            returnNormal(ControlFlowOps.invokeDynamicFunctionIn("fwdGetModules", fwdCustomApis));
         }
         else
         {
            returnNormal("");
         }
      }));
   }

   public void inMsgFwdMenuItemSelected(final character _hch_1)
   {
      character hch_1 = UndoableFactory.initInput(_hch_1);
      handle mbhandle = UndoableFactory.handle();
      character evt = UndoableFactory.character();
      handle h_5 = UndoableFactory.handle();
      
      internalProcedure(new Block((Body) () -> 
      {
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(lastWindow)
         }).first());
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            returnNormal();
         }
         
         mbhandle.assign(fwdEmbeddedWindow.getMenubar());
         
         if (_not(mbhandle.isValid()))
         {
            returnNormal();
         }
         
         h_5.assign(handle.fromString(hch_1));
         
         if (_or(or(not(h_5.isValid()), () -> not(HandleOps.canQuery(h_5, "TYPE"))), () -> not(hasWidget(mbhandle, h_5))))
         {
            returnNormal();
         }
         
         /*
            events:
            menu-drop: for sub-menu - only if it has children
            value-changed : for toggle-box
            choose : for everything else
            */
         if (_isEqual(h_5.unwrapType().getResourceType(), "SUB-MENU"))
         {
            if (_isUnknown(h_5.unwrapTree().firstChild()))
            {
               returnNormal();
            }
            
            evt.assign("menu-drop");
         }
         else
         {
            if (_and(HandleOps.canQuery(h_5, "toggle-box"), () -> h_5.unwrapMenuItem().isToggleBox()))
            {
               evt.assign("value-changed");
            }
            else
            {
               evt.assign("choose");
            }
         }
         apply(h_5, evt);
      }));
   }

   public void inMsgFwdToolbarItemSelected(final character _hch_1)
   {
      character hch_1 = UndoableFactory.initInput(_hch_1);
      handle tbhandle = UndoableFactory.handle();
      handle h_5 = UndoableFactory.handle();
      
      internalProcedure(new Block((Body) () -> 
      {
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(lastWindow)
         }).first());
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            returnNormal();
         }
         
         tbhandle.assign(fwdEmbeddedWindow.getToolbar());
         
         if (_not(tbhandle.isValid()))
         {
            returnNormal();
         }
         
         /*
            events:
            choose: for button click
            */
         h_5.assign(handle.fromString(hch_1));
         
         if (_not(h_5.isValid()))
         {
            returnNormal();
         }
         
         if (_not(new FindQuery(pageToolbar, "upper(pageToolbar.pname) = ? and pageToolbar.pnum = ? and pageToolbar.button = ?", null, "pageToolbar.id asc", new Object[]
         {
            toUpperCase(lastWindow),
            lastPage,
            h_5
         }, LockType.NONE).hasOne()))
         {
            returnNormal();
         }
         
         apply(h_5, "choose");
      }));
   }

   public void inMsgFwdGetProgramModule(final character _module, final character programName_1)
   {
      character module = UndoableFactory.initInput(_module);
      
      internalProcedure(new Block((Init) () -> 
      {
         UndoableFactory.initOutput(programName_1);
      }, 
      (Body) () -> 
      {
         if (_and(fwdCustomApis.isValid(), () -> matchesList(fwdCustomApis.unwrapProcedure().internalEntries(), "fwdGetProgramModule")))
         {
            ControlFlowOps.invokeInWithMode("fwdGetProgramModule", fwdCustomApis, "IO", module, programName_1);
            programName_1.assign(resolveProgramName(programName_1));
         }
         else
         {
            programName_1.assign(module);
         }
      }));
   }

   public void inMsgFwdDestroyWindow(final integer _windowId)
   {
      integer windowId = UndoableFactory.initInput(_windowId);
      handle phandle = UndoableFactory.handle();
      
      internalProcedure(new Block((Body) () -> 
      {
         if (_and(notUnknown(windowId), () -> isNotEqual(windowId, lastWindowId)))
         {
            returnNormal();
         }
         
         doBlock(TransactionType.FULL, "blockLabel1", new Block((Body) () -> 
         {
            silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
            {
               toUpperCase(lastWindow)
            }).first());
            
            if (_not(fwdEmbeddedWindow.available()))
            {
               returnNormal();
            }
            
            pageToolbar.delete("upper(pname) = ?", toUpperCase(lastWindow));
            phandle.assign(fwdEmbeddedWindow.getPhandle());
            
            if ((phandle.isValid()).booleanValue())
            {
               OnPhrase[] onPhrase0 = new OnPhrase[]
               {
                  new OnPhrase(Condition.STOP, Action.LEAVE, "blockLabel2")
               };
               
               doBlock(TransactionType.SUB, "blockLabel2", onPhrase0, new Block((Body) () -> 
               {
                  if ((fwdEmbeddedWindow.isIsadm2()).booleanValue())
                  {
                     silent(() -> ControlFlowOps.invokeIn("destroyObject", phandle));
                  }
                  else
                  {
                     silent(() -> ControlFlowOps.invokeInWithMode("dispatch", phandle, "I", new character("destroy")));
                  }
                  
                  silent(() -> HandleOps.delete(phandle));
               }));
            }
            fwdEmbeddedWindow.deleteRecord();
            lastWindowId.setUnknown();
            lastWindow.setUnknown();
            lastPage.setUnknown();
         }));
      }));
   }

   public void inMsgFwdSwitchWindow(final integer _windowId, final integer _pageNum, final character windowProgram, final character windowTitle)
   {
      integer windowId = UndoableFactory.initInput(_windowId);
      integer pageNum = UndoableFactory.initInput(_pageNum);
      handle hwin = UndoableFactory.handle();
      handle hframe = UndoableFactory.handle();
      character admVersion = UndoableFactory.character();
      character pages = UndoableFactory.character();
      handle phandle = UndoableFactory.handle();
      handle tbhandle = UndoableFactory.handle();
      handle htab = UndoableFactory.handle();
      
      internalProcedure(new Block((Init) () -> 
      {
         UndoableFactory.initOutput(windowProgram);
         UndoableFactory.initOutput(windowTitle);
      }, 
      (Body) () -> 
      {
         if (_isEqual(lastWindowId, windowId))
         {
            windowProgram.assign(lastWindow);
            lastPage.assign(pageNum);
            hwin.assign(LogicalTerminal.getWidgetHandleById(windowId));
            windowTitle.assign(hwin.unwrapWidget().getTitle());
            returnNormal();
         }
         else
         {
            /* check the previous window - if it has been destroyed, remove it.*/
            hwin.assign(LogicalTerminal.getWidgetHandleById(lastWindowId));
            
            if (_not(hwin.isValid()))
            {
               ControlFlowOps.invokeWithMode("inMsg_fwdDestroyWindow", "I", lastWindowId);
            }
            
            hwin.setUnknown();
         }
         hwin.assign(LogicalTerminal.getWidgetHandleById(windowId));
         
         if (_not(hwin.isValid()))
         {
            /* this window was already destroyed/not val*/
            windowProgram.setUnknown();
            windowTitle.setUnknown();
            returnNormal();
         }
         
         phandle.assign(hwin.unwrapInstantiatingProcedure().instantiatingProcedure());
         silent(() -> ControlFlowOps.invokeWithMode("getFwdHandles", "OOOOI", hwin, hframe, admVersion, pages, phandle));
         
         if (_or(or(or(not(hwin.isValid()), () -> isNotEqual(hwin.unwrapWidget().getHWND(), windowId)), () -> isUnknown(admVersion)), () -> isUnknown(hframe)))
         {
            /* this window was not executed from an ADM-style program.*/
            windowProgram.setUnknown();
            windowTitle.setUnknown();
            returnNormal();
         }
         
         windowProgram.assign(resolveProgramName(new character(phandle.unwrap().name())));
         windowTitle.assign(hwin.unwrapWidget().getTitle());
         /* set this now, as later inMsg_fwdShowWindow will be called and we need to be set on th*/
         lastWindow.assign(windowProgram);
         lastWindowId.assign(windowId);
         lastPage.assign(pageNum);
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(windowProgram)
         }).first());
         
         if (_and(fwdEmbeddedWindow.available(), () -> not(fwdEmbeddedWindow.getPhandle().isValid())))
         {
            doBlock(TransactionType.FULL, "blockLabel3", new Block((Body) () -> 
            {
               fwdEmbeddedWindow.deleteRecord();
            }));
         }
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            doBlock(TransactionType.FULL, "blockLabel4", new Block((Body) () -> 
            {
               if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetToolbarFrame")).booleanValue())
               {
                  tbhandle.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetToolbarFrame", phandle));
               }
               else
               {
                  tbhandle.setUnknown();
               }
               
               /* do this before anything else */
               fwdEmbeddedWindow.create();
               fwdEmbeddedWindow.setPname(windowProgram);
               fwdEmbeddedWindow.setPhandle(phandle);
               fwdEmbeddedWindow.setMenubar(hwin.unwrapWindow().getMenubar());
               fwdEmbeddedWindow.setToolbar(tbhandle);
               fwdEmbeddedWindow.setIsadm2(isGreaterThanOrEqual(admVersion, "ADM2.2"));
               fwdEmbeddedWindow.setPages(new logical(false));
               ControlFlowOps.invokeWithMode("initEmbeddedWindow", "I", hwin);
               
               if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetTabHandle")).booleanValue())
               {
                  htab.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetTabHandle", phandle));
               }
               else
               {
                  htab.setUnknown();
               }
               
               if ((htab.isValid()).booleanValue())
               {
                  NamedEventManager.unsubscribe(htab, "changeFolderPage");
               }
            }));
         }
         else
         {
            hwin.unwrapWidget().setVisible(new logical(true));
            hwin.unwrapWidget().moveToTop();
         }
      }));
   }

   public void inMsgFwdShowWindow(final character _pfile, final integer _pnum, final character _module, final integer _bgColorRgb, final integer windowId, final character pages)
   {
      character pfile = UndoableFactory.initInput(_pfile);
      integer pnum = UndoableFactory.initInput(_pnum);
      character module = UndoableFactory.initInput(_module);
      integer bgColorRgb = UndoableFactory.initInput(_bgColorRgb);
      character admVersion = UndoableFactory.character();
      handle phandle = UndoableFactory.handle();
      handle hframe = UndoableFactory.handle();
      handle tbhandle = UndoableFactory.handle();
      handle hwin = UndoableFactory.handle();
      handle htab = UndoableFactory.handle();
      handle lastWin = UndoableFactory.handle();
      
      internalProcedure(new Block((Init) () -> 
      {
         UndoableFactory.initOutput(windowId);
         UndoableFactory.initOutput(pages);
      }, 
      (Body) () -> 
      {
         windowId.setUnknown();
         pages.setUnknown();
         /* standardize the program file na*/
         pfile.assign(resolveProgramName(pfile));
         /*
            do not destroy the window, just create a new module and select it.
            if lastWindow <> pfile 
            then do:
               run inMsg_fwdDestroyWindow(lastWindowId).
            end.
            */
         lastWindow.assign(pfile);
         lastPage.assign(pnum);
         silent(() -> new FindQuery(fwdEmbeddedWindow, "upper(fwdEmbeddedWindow.pname) = ?", null, "fwdEmbeddedWindow.id asc", new Object[]
         {
            toUpperCase(pfile)
         }).first());
         
         if (_and(fwdEmbeddedWindow.available(), () -> not(fwdEmbeddedWindow.getPhandle().isValid())))
         {
            doBlock(TransactionType.FULL, "blockLabel5", new Block((Body) () -> 
            {
               fwdEmbeddedWindow.deleteRecord();
            }));
         }
         
         if (_not(fwdEmbeddedWindow.available()))
         {
            doBlock(TransactionType.FULL, "blockLabel6", new Block((Body) () -> 
            {
               lastWin.assign(SessionUtils.lastChild());
               
               if ((fwdCustomApis.isValid()).booleanValue())
               {
                  if ((matchesList(fwdCustomApis.unwrapProcedure().internalEntries(), "fwdCheckAccess")).booleanValue())
                  {
                     if (_not(new logical(ControlFlowOps.invokeDynamicFunctionInWithMode("fwdCheckAccess", fwdCustomApis, "II", pfile, module))))
                     {
                        returnNormal();
                     }
                  }
                  else
                  {
                  }
                  
                  if ((matchesList(fwdCustomApis.unwrapProcedure().internalEntries(), "fwdRunModule")).booleanValue())
                  {
                     phandle.assign(ControlFlowOps.invokeDynamicFunctionInWithMode("fwdRunModule", fwdCustomApis, "II", pfile, module));
                  }
                  else
                  {
                     ControlFlowOps.invokePersistentSet((pfile).toStringMessage(), phandle);
                  }
               }
               else
               {
                  /* TODO: FWD security plugin */
                  ControlFlowOps.invokePersistentSet((pfile).toStringMessage(), phandle);
               }
               
               if (_not(phandle.isValid()))
               {
                  messageBox(new Object[]
                  {
                     "Module ",
                     module,
                     " can not be ran."
                  }, ALERT_ERROR, BTN_OK, (String) null);
                  returnNormal();
               }
               
               silent(() -> ControlFlowOps.invokeWithMode("getFwdHandles", "OOOOI", hwin, hframe, admVersion, pages, phandle));
               
               if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetTabHandle")).booleanValue())
               {
                  htab.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetTabHandle", phandle));
               }
               else
               {
                  htab.setUnknown();
               }
               
               /* TODO: what if the business logic starts its own event-processing loop (i.e. WAIT-FOR*/
               /* with its main ADM window? this will never be reached until that window is discarde*/
               if (_notUnknown(hwin))
               {
                  if (_or(not(hwin.isValid()), () -> and(and(notUnknown(pages), () -> isNotEqual(pages, "")), () -> not(htab.isValid()))))
                  {
                     /* if the tab handle was not initialized but this window has pages, it means th*/
                     /* module initialization/shown was aborted by some business logi*/
                     pages.setUnknown();
                     windowId.setUnknown();
                     silent(() -> HandleOps.delete(phandle));
                     returnNormal();
                  }
                  
                  windowId.assign(hwin.unwrapWidget().getHWND());
               }
               else
               {
                  /* if business logic aborted window initialization, do nothi*/
                  if (_isEqual(lastWin, SessionUtils.lastChild()))
                  {
                     pages.setUnknown();
                     windowId.setUnknown();
                     silent(() -> HandleOps.delete(phandle));
                     returnNormal();
                  }
                  
                  /* the windowId is needed, but there are cases where the window is opened by another 
                              program... use SESSION:LAST-CHILD instead to get last opened window. */
                  hwin.assign(SessionUtils.lastChild());
                  windowId.assign(hwin.unwrapWidget().getHWND());
                  lastWindowId.assign(windowId);
                  
                  if ((jsStatusArea).booleanValue())
                  {
                     hwin.unwrapWindow().setStatusArea(new logical(false));
                  }
                  
                  if ((jsMessageArea).booleanValue())
                  {
                     hwin.unwrapWindow().setMessageArea(new logical(false));
                  }
                  
                  if (_notUnknown(bgColorRgb))
                  {
                     hwin.unwrapWidget().setBgColorRgb(bgColorRgb);
                  }
                  
                  hwin.unwrapWidget().setVisible(new logical(true));
                  hwin.unwrapWidget().moveToTop();
                  hwin.setUnknown();
               }
               
               if (_or(or(isUnknown(admVersion), () -> isUnknown(hframe)), () -> isUnknown(hwin)))
               {
                  /* TODO: show some error message ? */
                  HandleOps.delete(phandle);
                  returnNormal();
               }
               
               hwin.unwrapDisableRedraw().setDisableRedraw(new logical(true));
               
               if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetToolbarFrame")).booleanValue())
               {
                  tbhandle.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetToolbarFrame", phandle));
               }
               else
               {
                  tbhandle.setUnknown();
               }
               
               /* do this before anything else */
               fwdEmbeddedWindow.create();
               fwdEmbeddedWindow.setPname(pfile);
               fwdEmbeddedWindow.setPhandle(phandle);
               fwdEmbeddedWindow.setMenubar(hwin.unwrapWindow().getMenubar());
               fwdEmbeddedWindow.setToolbar(tbhandle);
               fwdEmbeddedWindow.setIsadm2(isGreaterThanOrEqual(admVersion, "ADM2.2"));
               fwdEmbeddedWindow.setPages(new logical(false));
               ControlFlowOps.invokeWithMode("initEmbeddedWindow", "I", hwin);
               hwin.unwrapDisableRedraw().setDisableRedraw(new logical(true));
               
               if ((fwdEmbeddedWindow.isIsadm2()).booleanValue())
               {
                  ControlFlowOps.invokeIn("initializeObject", phandle);
               }
               else
               {
                  silent(() -> ControlFlowOps.invokeInWithMode("dispatch", phandle, "I", new character("initialize")));
               }
               
               if ((htab.isValid()).booleanValue())
               {
                  NamedEventManager.unsubscribe(htab, "changeFolderPage");
               }
            }));
         }
         else
         {
            phandle.assign(fwdEmbeddedWindow.getPhandle());
            tbhandle.assign(fwdEmbeddedWindow.getToolbar());
         }
         ControlFlowOps.invokeWithMode("initEmbeddedPage", "IIIIOO", phandle, pnum, new logical(true), bgColorRgb, windowId, pages);
      }));
   }

   private void initEmbeddedPage(final handle _phandle, final integer _pnum, final logical _selectPage, final integer _bgColorRgb, final integer windowId, final character pages)
   {
      handle phandle = UndoableFactory.initInput(_phandle);
      integer pnum = UndoableFactory.initInput(_pnum);
      logical selectPage = UndoableFactory.initInput(_selectPage);
      integer bgColorRgb = UndoableFactory.initInput(_bgColorRgb);
      character admVersion = UndoableFactory.character();
      handle hframe = UndoableFactory.handle();
      handle tbhandle = UndoableFactory.handle();
      handle hwin = UndoableFactory.handle();
      handle hwid_1 = UndoableFactory.handle();
      integer wx = UndoableFactory.integer();
      integer wy = UndoableFactory.integer();
      
      internalProcedure(new Block((Init) () -> 
      {
         UndoableFactory.initOutput(windowId);
         UndoableFactory.initOutput(pages);
      }, 
      (Body) () -> 
      {
         tbhandle.assign(fwdEmbeddedWindow.getToolbar());
         silent(() -> ControlFlowOps.invokeWithMode("getFwdHandles", "OOOOI", hwin, hframe, admVersion, pages, phandle));
         hwin.unwrapDisableRedraw().setDisableRedraw(new logical(true));
         wx.assign(divide(minus(LogicalTerminal.getWidthPixels(), hwin.unwrapSizeable().getWidthPixels()), 2));
         wy.assign(divide(minus(LogicalTerminal.getHeightPixels(), hwin.unwrapSizeable().getHeightPixels()), 2));
         wx.assign((_isLessThan(wx, 1) ? new integer(1) : wx));
         wy.assign((_isLessThan(wy, 1) ? new integer(1) : wy));
         
         if (_isNotEqual(hwin.unwrapXY().getX(), wx))
         {
            silent(() -> hwin.unwrapXY().setX(wx));
         }
         
         if (_isNotEqual(hwin.unwrapXY().getY(), wy))
         {
            silent(() -> hwin.unwrapXY().setY(wy));
         }
         
         if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetPageHandles")).booleanValue())
         {
            doBlock(TransactionType.FULL, "blockLabel7", new Block((Body) () -> 
            {
               if ((selectPage).booleanValue())
               {
                  fwdSelectPage.assign(true);
                  admSelectedPage.setUnknown();
                  
                  if ((fwdEmbeddedWindow.isIsadm2()).booleanValue())
                  {
                     silent(() -> ControlFlowOps.invokeInWithMode("selectPage", phandle, "I", pnum));
                  }
                  else
                  {
                     silent(() -> ControlFlowOps.invokeInWithMode("select-page", phandle, "I", pnum));
                  }
                  
                  if (_and(notUnknown(admSelectedPage), () -> isNotEqual(admSelectedPage, pnum)))
                  {
                     /* the ADM code has selected another page - switch to */
                     pnum.assign(admSelectedPage);
                     lastPage.assign(admSelectedPage);
                  }
                  else
                  {
                     if (_isUnknown(admSelectedPage))
                     {
                     }
                  }
                  admSelectedPage.setUnknown();
                  fwdSelectPage.assign(false);
               }
               
               if (_not(fwdEmbeddedWindow.isPages(minus((pnum), 1))))
               {
                  ControlFlowOps.invokeInWithMode("fwdGetPageHandles", phandle, "IOO", pnum, new ExtentExpr0(), numPages);
                  pageToolbar.delete("pnum = ? and upper(pname) = ?", pnum, toUpperCase(lastWindow));
                  
                  if (_and(notUnknown(tbhandle), () -> not(tbhandle.unwrapWidget().isVisible())))
                  {
                     tbhandle.unwrapWidget().setVisible(new logical(true));
                  }
                  
                  fwdEmbeddedWindow.setToolbars(minus((pnum), 1), listToolbar(tbhandle));
                  hwid_1.assign(hwin.unwrapTree().firstChild());
                  
                  LogicalOp whileClause0 = () -> notUnknown(hwid_1);
                  
                  doWhile("loopLabel2", whileClause0, new Block((Body) () -> 
                  {
                     if (_and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> not(or(isEqual(hwid_1, hframe), () -> isPageFrame(hwid_1)))))
                     {
                        hwid_1.unwrapWidget().setVisible(new logical(false));
                     }
                     
                     hwid_1.assign(hwid_1.unwrap().getNextSibling());
                  }));
                  
                  ControlFlowOps.invokeWithMode("initEmbeddedWidgets", "IIII", phandle, hwin, hframe, tbhandle);
                  fwdEmbeddedWindow.setPages(minus((pnum), 1), new logical(true));
               }
            }));
         }
         
         if ((jsStatusArea).booleanValue())
         {
            hwin.unwrapWindow().setStatusArea(new logical(false));
         }
         
         if ((jsMessageArea).booleanValue())
         {
            hwin.unwrapWindow().setMessageArea(new logical(false));
         }
         
         hwin.unwrapWindow().setWindowState(CommonWindow.WINDOW_STATE_NORMAL);
         hwin.unwrapWidget().setVisible(new logical(true));
         
         if (_notUnknown(bgColorRgb))
         {
            hwin.unwrapWidget().setBgColorRgb(bgColorRgb);
            
            if (_isEqual(hframe.unwrapWidget().getBgColor(), 15))
            {
               hframe.unwrapWidget().setBgColor(new integer());
            }
            
            hframe.unwrapWidget().setBgColorRgb(bgColorRgb);
         }
         hwin.unwrapDisableRedraw().setDisableRedraw(new logical(false));
         hwin.unwrapWidget().moveToTop();
         /* let the embedded JS driver know this window's ID, so that it can track it if it ge*/
         /* destroyed by the business log*/
         windowId.assign(hwin.unwrapWidget().getHWND());
         lastWindowId.assign(windowId);
         lastPage.assign(pnum);
      }));
   }

   private void getFwdHandles(final handle hwin, final handle hframe, final character admVersion, final character pages, final handle _phandle)
   {
      handle phandle = UndoableFactory.initInput(_phandle);
      
      internalProcedure(new Block((Init) () -> 
      {
         UndoableFactory.initOutput(hwin);
         UndoableFactory.initOutput(hframe);
         UndoableFactory.initOutput(admVersion);
         UndoableFactory.initOutput(pages);
      }, 
      (Body) () -> 
      {
         hwin.setUnknown();
         hframe.setUnknown();
         admVersion.setUnknown();
         
         if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetWindow")).booleanValue())
         {
            hwin.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetWindow", phandle));
         }
         
         if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetMainFrame")).booleanValue())
         {
            hframe.assign(ControlFlowOps.invokeDynamicFunctionIn(handle.class, "fwdGetMainFrame", phandle));
         }
         
         if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetADMVersion")).booleanValue())
         {
            admVersion.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "fwdGetADMVersion", phandle));
         }
         
         if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetPages")).booleanValue())
         {
            pages.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetPages", phandle));
         }
      }));
   }

   private void initEmbeddedWindow(final handle _hwin)
   {
      handle hwin = UndoableFactory.initInput(_hwin);
      
      internalProcedure(new Block((Body) () -> 
      {
         /* remove any menu attached to the window */
         hwin.unwrapWindow().setMenubar(new handle());
         /* remove status-area and message-area */
         silent(() -> hwin.unwrapWindow().setStatusArea(new logical(false)));
         silent(() -> hwin.unwrapWindow().setMessageArea(new logical(false)));
         hwin.unwrapTopOnly().setTopOnly(new logical(false));
         hwin.unwrapWindow().setAlwaysOnTop(new logical(false));
         hwin.unwrapWidget().setVisible(new logical(true));
         hwin.unwrapWidget().moveToTop();
      }));
   }

   private logical isPageFrame(final handle _hwid_1)
   {
      handle hwid_1 = TypeFactory.initInput(_hwid_1);
      integer i = UndoableFactory.integer();
      
      return function("isPageFrame", logical.class, new Block((Body) () -> 
      {
         loopLabel3:
         for (i.assign(1); _isLessThanOrEqual(i, numPages); i.increment())
         {
            if (_isEqual(subscript(hpages, i), hwid_1.unwrapInstantiatingProcedure().instantiatingProcedure()))
            {
               returnNormal(true);
            }
         }
         
         returnNormal(false);
      }));
   }

   private character resolveProgramName(final character _programName_1)
   {
      character programName_1 = TypeFactory.initInput(_programName_1);
      character prog2 = UndoableFactory.character();
      
      return function("resolveProgramName", character.class, new Block((Body) () -> 
      {
         programName_1.assign(FileSystemOps.searchPath(programName_1));
         
         if (_isEqual(lastIndexOf(programName_1, ".r"), minus(length(programName_1), 1)))
         l1:
         {
            programName_1.assign(substring(programName_1, 1, minus(length(programName_1), 2)));
            prog2.assign(FileSystemOps.searchPath(concat(programName_1, ".w")));
            
            if (_notUnknown(prog2))
            {
               programName_1.assign(prog2);
               break l1;
            }
            
            prog2.assign(FileSystemOps.searchPath(concat(programName_1, ".p")));
            
            if (_notUnknown(prog2))
            {
               programName_1.assign(prog2);
               break l1;
            }
            
            messageBox(new Object[]
            {
               "Module ",
               programName_1,
               " can not be ran."
            }, ALERT_ERROR, BTN_OK, (String) null);
            programName_1.setUnknown();
         }
         returnNormal(programName_1);
      }));
   }

   private decimal screenRow(final handle _hwid_1)
   {
      handle hwid_1 = TypeFactory.initInput(_hwid_1);
      decimal hrow = UndoableFactory.decimal(new decimal(0));
      
      return function("screenRow", decimal.class, new Block((Body) () -> 
      {
         loopLabel4:
         while (_notUnknown(hwid_1))
         {
            hrow.assign(plus(hrow, hwid_1.unwrapXY().getY()));
            
            if ((HandleOps.canQuery(hwid_1, "border-top-pixels")).booleanValue())
            {
               hrow.assign(plus(hrow, hwid_1.unwrapFrame().borderTopPixels()));
            }
            
            hwid_1.assign(hwid_1.unwrapWidget().getFrameHandle());
         }
         
         returnNormal(hrow);
      }));
   }

   private decimal screenCol(final handle _hwid_1)
   {
      handle hwid_1 = TypeFactory.initInput(_hwid_1);
      decimal hcol = UndoableFactory.decimal(new decimal(0));
      
      return function("screenCol", decimal.class, new Block((Body) () -> 
      {
         loopLabel5:
         while (_notUnknown(hwid_1))
         {
            hcol.assign(plus(hcol, hwid_1.unwrapXY().getX()));
            
            if ((HandleOps.canQuery(hwid_1, "border-left-pixels")).booleanValue())
            {
               hcol.assign(plus(hcol, hwid_1.unwrapFrame().borderLeftPixels()));
            }
            
            hwid_1.assign(hwid_1.unwrapWidget().getFrameHandle());
         }
         
         returnNormal(hcol);
      }));
   }

   private void initEmbeddedWidgets(final handle _phandle, final handle _hwin, final handle _hframe, final handle _htoolbar)
   {
      handle phandle = UndoableFactory.initInput(_phandle);
      handle hwin = UndoableFactory.initInput(_hwin);
      handle hframe = UndoableFactory.initInput(_hframe);
      handle htoolbar = UndoableFactory.initInput(_htoolbar);
      character hiddenWidgets = UndoableFactory.character();
      handle hwid_1 = UndoableFactory.handle();
      character widname = UndoableFactory.character();
      integer idx = UndoableFactory.integer();
      decimal k = UndoableFactory.decimal();
      decimal wrow = UndoableFactory.decimal();
      decimal wcol = UndoableFactory.decimal();
      decimal minRow = UndoableFactory.decimal(new decimal());
      decimal minCol = UndoableFactory.decimal(new decimal());
      
      internalProcedure(new Block((Body) () -> 
      {
         if ((matchesList(phandle.unwrapProcedure().internalEntries(), "fwdGetHiddenEmbeddedWidgets")).booleanValue())
         {
            hiddenWidgets.assign(ControlFlowOps.invokeDynamicFunctionIn("fwdGetHiddenEmbeddedWidgets", phandle));
         }
         else
         {
            hiddenWidgets.assign("");
         }
         
         /* hide all widgets which are not visible in embedded mode */
         idx.assign(1);
         
         loopLabel6:
         while (_isLessThanOrEqual(idx, numEntries(hiddenWidgets, ",")))
         {
            silent(() -> widname.assign(entry(idx, hiddenWidgets, ",")));
            
            if (_not(ErrorManager.isError()))
            {
               hwid_1.assign(hframe.unwrapTree().firstChild().unwrapTree().firstChild());
               
               loopLabel7:
               while (_notUnknown(hwid_1))
               {
                  if (_isEqual(hwid_1.unwrap().name(), widname))
                  {
                     hwid_1.unwrapWidget().setHidden(new logical(true));
                  }
                  
                  hwid_1.assign(hwid_1.unwrap().getNextSibling());
               }
            }
            idx.assign(plus(idx, 1));
         }
         
         hwid_1.assign(hframe.unwrapTree().firstChild().unwrapTree().firstChild());
         
         LogicalOp whileClause1 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel8", whileClause1, new Block((Body) () -> 
         {
            /* hwid:type = "FRAME" - this test was removed, as there may be non-FRAME widgets in t*/
            /* main frame, which must be hidd*/
            if (_or(not(isPageFrame(hwid_1)), () -> isAdditionalToolbar(htoolbar, hwid_1)))
            {
               hwid_1.unwrapWidget().setHidden(new logical(true));
            }
            
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         /* go through all widgets and compute min row/col from non-hidden ones */
         hwid_1.assign(hframe.unwrapTree().firstChild().unwrapTree().firstChild());
         
         LogicalOp whileClause2 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel9", whileClause2, new Block((Body) () -> 
         {
            if (_not(hwid_1.unwrapWidget().isHidden()))
            {
               wrow.assign(screenRow(hwid_1));
               wcol.assign(screenCol(hwid_1));
               
               if (_or(isUnknown(minRow), () -> isGreaterThan(minRow, wrow)))
               {
                  minRow.assign(wrow);
               }
               
               if (_or(isUnknown(minCol), () -> isGreaterThan(minCol, wcol)))
               {
                  minCol.assign(wcol);
               }
            }
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         /* go through all page frames in the window and compute min row/col */
         hwid_1.assign(hwin.unwrapTree().firstChild());
         
         LogicalOp whileClause3 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel10", whileClause3, new Block((Body) () -> 
         {
            if (_and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> isPageFrame(hwid_1)))
            {
               wrow.assign(screenRow(hwid_1));
               wcol.assign(screenCol(hwid_1));
               
               if (_or(isUnknown(minRow), () -> isGreaterThan(minRow, wrow)))
               {
                  minRow.assign(wrow);
               }
               
               if (_or(isUnknown(minCol), () -> isGreaterThan(minCol, wcol)))
               {
                  minCol.assign(wcol);
               }
            }
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         /* re-adjust the widget location in the main frame */
         hwid_1.assign(hframe.unwrapTree().firstChild().unwrapTree().firstChild());
         
         LogicalOp whileClause4 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel11", whileClause4, new Block((Body) () -> 
         {
            if (_not(hwid_1.unwrapWidget().isHidden()))
            {
               wrow.assign(screenRow(hwid_1));
               wcol.assign(screenCol(hwid_1));
               k.assign(minus(wrow, minus(minRow, 1)));
               
               if (_isNotEqual(hwid_1.unwrapXY().getY(), k))
               {
                  silent(() -> hwid_1.unwrapXY().setY(new integer(k)));
               }
            }
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         /* re-adjust the location for non-nested page frames, relative to the widgets in hframe */
         hwid_1.assign(hwin.unwrapTree().firstChild());
         
         LogicalOp whileClause5 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel12", whileClause5, new Block((Body) () -> 
         {
            if (_and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> isPageFrame(hwid_1)))
            {
               wrow.assign(screenRow(hwid_1));
               wcol.assign(screenCol(hwid_1));
               k.assign(minus(wrow, minus(minRow, 1)));
               
               if (_isNotEqual(hwid_1.unwrapXY().getY(), k))
               {
                  silent(() -> hwid_1.unwrapXY().setY(new integer(k)));
               }
            }
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         /* one more time, to move the main frame and non-nested page frames */
         minRow.setUnknown();
         minCol.setUnknown();
         /* go through all page frames in the window and compute min row/col */
         hwid_1.assign(hwin.unwrapTree().firstChild());
         
         LogicalOp whileClause6 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel13", whileClause6, new Block((Body) () -> 
         {
            if (_and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> or(isEqual(hwid_1, hframe), () -> isPageFrame(hwid_1))))
            {
               wrow.assign(screenRow(hwid_1));
               wcol.assign(screenCol(hwid_1));
               
               if (_or(isUnknown(minRow), () -> isGreaterThan(minRow, wrow)))
               {
                  minRow.assign(wrow);
               }
               
               if (_or(isUnknown(minCol), () -> isGreaterThan(minCol, wcol)))
               {
                  minCol.assign(wcol);
               }
            }
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         /* re-adjust the location for non-nested page frames, relative to main frame */
         hwid_1.assign(hwin.unwrapTree().firstChild());
         
         LogicalOp whileClause7 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel14", whileClause7, new Block((Body) () -> 
         {
            if (_and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> or(isEqual(hwid_1, hframe), () -> isPageFrame(hwid_1))))
            {
               wrow.assign(screenRow(hwid_1));
               wcol.assign(screenCol(hwid_1));
               k.assign(minus(wrow, minus(minRow, 1)));
               
               if (_isNotEqual(hwid_1.unwrapXY().getY(), k))
               {
                  silent(() -> hwid_1.unwrapXY().setY(new integer(k)));
               }
            }
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
      }));
   }

   private character listMenuTree(final handle _h_5, final integer _lvl_1)
   {
      handle h_5 = TypeFactory.initInput(_h_5);
      integer lvl_1 = TypeFactory.initInput(_lvl_1);
      /*
      {
         "sensitive" : true/false, // if set, all menu-items are disabled (only for root)
         "visible"   : true/false, // if set, menubar is not visible (only for root)
         "menu-items" :
         [ 
            {
               "handle"      : "",
               "type"        : "MENU-ITEM/SUB-MENU",
               "subtype"     : "NORMAL/RULE/SKIP", // only for type=MENU-ITEM, default NORMAL
               "label"       : "", // this is the preprocessed 'label' attribute, with mnemonic resolved
               "accelerator" : "", // optional
               "mnemonic"    : 0,1,2,etc, // optional, default 0; this is the resolved mnemonic index, in the preprocessed label
               "read-only"   : true/false, // optional, default false; can't be chosen (same as sensitive=false?)
               "sensitive"   : true/false, // optional, default true
               // "visible"     : true/false, // optional, default true - not used, does nothing for menu-related stuff
               "toggle-box"  : true/false,  // optional, default false
               "checked"     : true/false, // optional, default false,  only for toggle-box
               "menu-items"  : // optional, only for sub-menus
               [
                  {
                     // recursive, again each element
                  }
               ]
            },
            ...
         ]
      }
      */
      character res = UndoableFactory.character();
      handle hch_1 = UndoableFactory.handle();
      
      return function("list-menu-tree", character.class, new Block((Body) () -> 
      {
         res.assign("");
         res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"menu-items\" : \n"));
         res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "[\n"));
         lvl_1.assign(plus(lvl_1, 1));
         hch_1.assign(h_5.unwrapTree().firstChild());
         
         LogicalOp whileClause8 = () -> notUnknown(hch_1);
         
         doWhile("loopLabel15", whileClause8, new Block((Body) () -> 
         {
            res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "{\n"));
            lvl_1.assign(plus(lvl_1, 1));
            res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"handle\"      : \"", valueOf(hch_1), "\",\n"));
            res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"type\"        : \"", hch_1.unwrapType().getResourceType(), "\",\n"));
            
            if (_and(and(HandleOps.canQuery(hch_1, "subtype"), () -> notUnknown(hch_1.unwrapSubType().getSubType())), () -> isNotEqual(hch_1.unwrapSubType().getSubType(), "NORMAL")))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"subtype\"     : \"", hch_1.unwrapSubType().getSubType(), "\",\n"));
            }
            
            if (_and(HandleOps.canQuery(hch_1, "preprocessed-label"), () -> notUnknown(hch_1.unwrapMnemonic().getPreprocessedLabel())))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"label\"       : \"", hch_1.unwrapMnemonic().getPreprocessedLabel(), "\",\n"));
            }
            
            if (_and(HandleOps.canQuery(hch_1, "accelerator"), () -> notUnknown(hch_1.unwrapMenuItem().getAccelerator())))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"accelerator\" : \"", hch_1.unwrapMenuItem().getAccelerator(), "\",\n"));
            }
            
            if (_and(and(HandleOps.canQuery(hch_1, "mnemonic"), () -> notUnknown(hch_1.unwrapMnemonic().getMnemonic())), () -> isNotEqual(hch_1.unwrapMnemonic().getMnemonic(), 0)))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"mnemonic\"    : \"", valueOf(hch_1.unwrapMnemonic().getMnemonic()), "\",\n"));
            }
            
            if (_and(and(HandleOps.canQuery(hch_1, "sensitive"), () -> notUnknown(hch_1.unwrapSensitive().isSensitive())), () -> not(hch_1.unwrapSensitive().isSensitive())))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"sensitive\"   : ", valueOf(hch_1.unwrapSensitive().isSensitive(), "true/false"), ",\n"));
            }
            
            if (_and(and(HandleOps.canQuery(hch_1, "read-only"), () -> notUnknown(hch_1.unwrapWriteProtectable().isReadOnly())), () -> hch_1.unwrapWriteProtectable().isReadOnly()))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"read-only\"   : ", valueOf(hch_1.unwrapWriteProtectable().isReadOnly(), "true/false"), ",\n"));
            }
            
            /*
                  if (can-query(hch, "visible") and hch:visible <> ? and not hch:visible)
                  then do:
                     res = res + fill(" " , lvl * 3) + '"visible"     : ' + string(hch:visible, "true/false") + ',
            '.
                  end.
                  */
            if (_and(and(HandleOps.canQuery(hch_1, "toggle-box"), () -> notUnknown(hch_1.unwrapMenuItem().isToggleBox())), () -> hch_1.unwrapMenuItem().isToggleBox()))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"toggle-box\"  : ", valueOf(hch_1.unwrapMenuItem().isToggleBox(), "true/false"), ",\n"));
            }
            
            if (_and(and(HandleOps.canQuery(hch_1, "checked"), () -> notUnknown(hch_1.unwrapWidget().isChecked())), () -> hch_1.unwrapWidget().isChecked()))
            {
               res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "\"checked\"     : ", valueOf(hch_1.unwrapWidget().isChecked(), "true/false"), ",\n"));
            }
            
            if ((HandleOps.canQuery(hch_1, "first-child")).booleanValue())
            {
               res.assign(concat(res, listMenuTree(hch_1, lvl_1)));
            }
            
            res.assign(trim(res));
            
            if (_isEqual(lastIndexOf(res, ","), length(res)))
            {
               res.assign(substring(res, 1, minus(length(res), 1)));
            }
            
            lvl_1.assign(minus(lvl_1, 1));
            res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "}", (_notUnknown(hch_1.unwrap().getNextSibling()) ? new character(",") : new character("")), "\n"));
            hch_1.assign(hch_1.unwrap().getNextSibling());
         }));
         
         lvl_1.assign(minus(lvl_1, 1));
         res.assign(concat(res, fill(" ", multiply(lvl_1, 3)), "]\n"));
         returnNormal(res);
      }));
   }

   private character listMenu(final handle _h_5)
   {
      handle h_5 = TypeFactory.initInput(_h_5);
      character res = UndoableFactory.character();
      
      return function("list-menu", character.class, new Block((Body) () -> 
      {
         res.assign("");
         res.assign(concat(res, "{\n"));
         res.assign(concat(res, fill(" ", 3), "\"sensitive\" : ", valueOf(h_5.unwrapSensitive().isSensitive(), "true/false"), ",\n"));
         res.assign(concat(res, fill(" ", 3), "\"visible\" : ", valueOf(h_5.unwrapWidget().isVisible(), "true/false"), ",\n"));
         res.assign(concat(res, listMenuTree(h_5, new integer(1))));
         res.assign(concat(res, "}\n"));
         returnNormal(res);
      }));
   }

   private character listToolbarButton(final handle _h_5, final integer _hx_1, final integer _hy_1)
   {
      handle h_5 = TypeFactory.initInput(_h_5);
      integer hx_1 = TypeFactory.initInput(_hx_1);
      integer hy_1 = TypeFactory.initInput(_hy_1);
      character res = UndoableFactory.character();
      integer x = UndoableFactory.integer();
      integer y = UndoableFactory.integer();
      
      return function("list-toolbar-button", character.class, new Block((Body) () -> 
      {
         if (_or(not(h_5.unwrapWidget().isVisible()), () -> and(not(h_5.unwrapWidget().getFrameHandle().unwrapWidget().isVisible()), () -> not(isPageFrame(h_5.unwrapWidget().getFrameHandle())))))
         {
            returnNormal("");
         }
         
         pageToolbar.create();
         pageToolbar.setPname(lastWindow);
         pageToolbar.setPnum(lastPage);
         pageToolbar.setButton(h_5);
         x.assign(plus(h_5.unwrapXY().getX(), hx_1));
         y.assign(plus(h_5.unwrapXY().getY(), hy_1));
         res.assign("");
         res.assign(concat(res, "   ", "{\n"));
         res.assign(concat(res, "   ", "\"handle\": \"", valueOf(h_5), "\",\n"));
         
         if (_notUnknown(h_5.unwrap().name()))
         {
            res.assign(concat(res, "      ", "\"name\": \"", h_5.unwrap().name(), "\",\n"));
         }
         
         if (_notUnknown(h_5.unwrapCommonField().getLabel()))
         {
            res.assign(concat(res, "      ", "\"label\": \"", h_5.unwrapCommonField().getLabel(), "\",\n"));
         }
         
         if (_notUnknown(h_5.unwrapWidget().getTooltip()))
         {
            res.assign(concat(res, "      ", "\"tooltip\": \"", h_5.unwrapWidget().getTooltip(), "\",\n"));
         }
         
         if (_and(isNotEqual(h_5.unwrapButton().getImageUp(), ""), () -> notUnknown(h_5.unwrapButton().getImageUp())))
         {
            res.assign(concat(res, "      ", "\"image\": \"", h_5.unwrapButton().getImageUp(), "\",\n"));
         }
         
         res.assign(concat(res, "      ", "\"sensitive\":", valueOf(h_5.unwrapSensitive().isSensitive(), "true/false"), ",\n"));
         res.assign(concat(res, "      ", "\"width\":", valueOf(h_5.unwrapSizeable().getWidthPixels(), ">>>>9"), ",\n"));
         res.assign(concat(res, "      ", "\"height\":", valueOf(h_5.unwrapSizeable().getHeightPixels(), ">>>>9"), ",\n"));
         res.assign(concat(res, "      ", "\"x\":", valueOf(x, ">>>>>9"), ",\n"));
         res.assign(concat(res, "      ", "\"y\":", valueOf(y, ">>>>>9"), "\n"));
         res.assign(concat(res, "   ", "},\n"));
         returnNormal(res);
      }));
   }

   private character listToolbarButtons(final handle _h_5, final integer _hx_1, final integer _hy_1)
   {
      handle h_5 = TypeFactory.initInput(_h_5);
      integer hx_1 = TypeFactory.initInput(_hx_1);
      integer hy_1 = TypeFactory.initInput(_hy_1);
      character res = UndoableFactory.character();
      handle hch_1 = UndoableFactory.handle();
      
      return function("list-toolbar-buttons", character.class, new Block((Body) () -> 
      {
         res.assign("");
         hch_1.assign(h_5.unwrapTree().firstChild());
         
         LogicalOp whileClause9 = () -> notUnknown(hch_1);
         
         doWhile("loopLabel16", whileClause9, new Block((Body) () -> 
         {
            if (_and(HandleOps.canQuery(hch_1, "first-child"), () -> notUnknown(hch_1.unwrapTree().firstChild())))
            {
               res.assign(concat(res, listToolbarButtons(hch_1, new integer(plus(hx_1, hch_1.unwrapXY().getX())), new integer(plus(hy_1, h_5.unwrapXY().getY())))));
            }
            else
            {
               if (_isEqual(hch_1.unwrapType().getResourceType(), "BUTTON"))
               {
                  res.assign(concat(res, listToolbarButton(hch_1, hx_1, hy_1)));
               }
            }
            hch_1.assign(hch_1.unwrap().getNextSibling());
         }));
         
         returnNormal(res);
      }));
   }

   private character listToolbar(final handle _h_5)
   {
      handle h_5 = TypeFactory.initInput(_h_5);
      character res = UndoableFactory.character();
      handle hwid_1 = UndoableFactory.handle();
      integer hpx = UndoableFactory.integer();
      integer hpy = UndoableFactory.integer();
      
      return function("list-toolbar", character.class, new Block((Body) () -> 
      {
         if (_not(h_5.isValid()))
         {
            returnNormal("");
         }
         
         res.assign("");
         res.assign(concat(res, "[\n"));
         res.assign(concat(res, listToolbarButtons(h_5, new integer(0), new integer(0))));
         
         if (_isEqual(h_5.unwrapWidget().getParentHandle().unwrapType().getResourceType(), "WINDOW"))
         {
            hpx.assign(0);
            hpy.assign(0);
         }
         else
         {
            hpx.assign(h_5.unwrapWidget().getParentHandle().unwrapWidget().getParentHandle().unwrapXY().getX());
            hpy.assign(h_5.unwrapWidget().getParentHandle().unwrapWidget().getParentHandle().unwrapXY().getY());
         }
         
         hwid_1.assign(h_5.unwrapWidget().getParentHandle().unwrapTree().firstChild());
         
         LogicalOp whileClause10 = () -> notUnknown(hwid_1);
         
         doWhile("loopLabel17", whileClause10, new Block((Body) () -> 
         {
            if (_and(and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> isNotEqual(hwid_1, h_5)), () -> isEqual(hwid_1.unwrapXY().getY(), h_5.unwrapXY().getY())))
            {
               res.assign(concat(res, listToolbarButtons(hwid_1, new integer(plus(hpx, hwid_1.unwrapXY().getX())), new integer(plus(hpy, hwid_1.unwrapXY().getY())))));
            }
            
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }));
         
         res.assign(trim(res));
         
         if (_isEqual(lastIndexOf(res, ","), length(res)))
         {
            res.assign(substring(res, 1, minus(length(res), 1)));
         }
         
         res.assign(concat(res, "]\n"));
         returnNormal(res);
      }));
   }

   private logical isAdditionalToolbar(final handle _h_5, final handle _hch_1)
   {
      handle h_5 = TypeFactory.initInput(_h_5);
      handle hch_1 = TypeFactory.initInput(_hch_1);
      handle hwid_1 = UndoableFactory.handle();
      
      return function("is-additional-toolbar", logical.class, new Block((Body) () -> 
      {
         if (_or(not(h_5.isValid()), () -> isNotEqual(hch_1.unwrapXY().getY(), h_5.unwrapXY().getY())))
         {
            returnNormal(false);
         }
         
         hwid_1.assign(h_5.unwrapWidget().getParentHandle().unwrapTree().firstChild());
         
         loopLabel18:
         while (_notUnknown(hwid_1))
         {
            if (_and(and(and(isEqual(hwid_1.unwrapType().getResourceType(), "FRAME"), () -> isNotEqual(hwid_1, h_5)), () -> isEqual(hwid_1, hch_1)), () -> isEqual(hwid_1.unwrapXY().getY(), h_5.unwrapXY().getY())))
            {
               returnNormal(true);
            }
            
            hwid_1.assign(hwid_1.unwrap().getNextSibling());
         }
         
         returnNormal(false);
      }));
   }

   private logical hasWidget(final handle _hparent_1, final handle _hchild_1)
   {
      handle hparent_1 = TypeFactory.initInput(_hparent_1);
      handle hchild_1 = TypeFactory.initInput(_hchild_1);
      handle hch_1 = UndoableFactory.handle();
      
      return function("has-widget", logical.class, new Block((Body) () -> 
      {
         hch_1.assign(hparent_1.unwrapTree().firstChild());
         
         LogicalOp whileClause11 = () -> notUnknown(hch_1);
         
         doWhile("loopLabel19", whileClause11, new Block((Body) () -> 
         {
            if (_isEqual(hch_1, hchild_1))
            {
               returnNormal(true);
            }
            
            if (_and(HandleOps.canQuery(hch_1, "first-child"), () -> notUnknown(hch_1.unwrapTree().firstChild())))
            {
               if ((hasWidget(hch_1, hchild_1)).booleanValue())
               {
                  returnNormal(true);
               }
            }
            hch_1.assign(hch_1.unwrap().getNextSibling());
         }));
         
         returnNormal(false);
      }));
   }

   public character fwdGetAdmversion()
   {
      return function("fwdGetADMVersion", character.class, new Block((Body) () -> 
      {
         returnNormal("ADM2.2");
      }));
   }

   private class ExtentExpr0
   extends OutputExtentParameter<handle>
   {
      public handle[] getVariable()
      {
         return hpages;
      }

      public void setVariable(final handle[] newRef)
      {
         hpages = newRef;
      }
   }
}
