/*--------------------------------------------------------------------------
    File        : Appserver.java
    Purpose     : Super procedure for appserver class.
                  Implements methods to start/restart and communicate with a  
                  server part of an object.                       
    Syntax      : RUN start-super-proc("adm2/appserver.p":U).
    
    Author      : Roland de Pijper, Progress Software BV
    Modified    : 17/11/200
    Modified    : 01/24/2001  Progress Software Corp. 
                  Made part of 9.1C ADM2 (inherited by data and sbo classes) with
                  the following changes.  
                  - Removed use of connectServerObject.p in anticipation of 
                    ICF session and context services... A runServerProcedure 
                    function was implemented to simplify an override to achieve 
                    this functionality. 
                  - Minor adjustments of some comments and logic from equivalent
                    methods in data and sbo classes.
                  - Added unbindServer and bindserver (from sbo class 9.1B05)
                    and removed setASBound.    
                  - Added stateless client 'events' initializeServerObject
                    and destroyServerObject to simplify overrides for 
                    objects that needs context management and also to make 
                    a call to initializeObject on the server as default as 
                    this is a current adm 'rule'. A ASInitializeOnRun property
                    that defaults to yes, decides whether initializeServerObject
                    is run.         
                  - destroyObject -> disconnectObject (both from data class)
                  - Added a startServerObject that encapsulates logic from 
                    initializeObject in data and sbo classes. A new logical 
                    AsHasStarted property decides whether to start or 
                    restart from getASHandle. Also added a runServerObject 
                    procedure that is called from start- and restartServerObject
                    and has (most of) the common start/restart code.                                                             
    Notes:        This class does not directly support context management,
                  but its two stateless client 'events'; initializeServerObject 
                  and destroyServerObject is the 'events' that implementors 
                  overrides for context management on the client. 
               -  This class does NOT show any messages.   
                  Errors hardcoded messages to addMessage. The methods also
                  returns ERROR 'ADM-ERROR' on an error, but this is just for 
                  direct calls and are NOT returned up the stack.        
                  Implementors can override getAsHandle and/or start-, 
                  restartServerProcedure to check for anyMessage and show 
                  messages.     
  ------------------------------------------------------------------------*/

package com.goldencode.hotel.common.adm2;

import com.goldencode.p2j.util.*;
import com.goldencode.p2j.ui.*;
import com.goldencode.util.*;

import static com.goldencode.p2j.util.BlockManager.*;
import static com.goldencode.p2j.util.CompareOps.*;
import static com.goldencode.p2j.util.logical.*;
import static com.goldencode.p2j.util.TextOps.*;
import static com.goldencode.p2j.util.ProcedureManager.*;
import static com.goldencode.p2j.util.character.*;
import static com.goldencode.p2j.ui.LogicalTerminal.*;
import static com.goldencode.p2j.util.ErrorManager.*;

/**
 * Business logic (converted to Java from the 4GL source code
 * in common/adm2/appserver.p).
 */
public class Appserver
{
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*          This .W file was created with the Progress UIB.             */   
   /*----------------------------------------------------------------------*/   
   /* ***************************  Definitions  ************************** */   
   /* Custom exclude file */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*--------------------------------------------------------------------------
       File        : appserverexclcustom.i
       Purpose     : Defines EXCLUDE-xxx preprocessor variables to exclude
                     functions or internal procedures in the super procedure
                     where it is referenced.
       Syntax      : {src/adm2/custom/appserverexclcustom.i}
   
       Modified    : 01/24/2001
       
       Notes       : Referenced in src/adm2/appserver.p
     ------------------------------------------------------------------------*/   
   /* Include the file which defines AppServerConnect procedures. */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /* adecomm/appserv.i checks to see if the persistent procedure adecomm/as-utils.w
                        has been started up, and if not, it starts it and then makes
                        is a SUPER procedure of the calling procedure.
                        
      Created: 11/26/97 by Ross Hunter
                                                                                    */   
   handle appSrvUtils = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "appsrvutils", true, new handle());

   /* _UIB-CODE-BLOCK-END */   
   /* *********************** Procedure Settings ************************ */   
   /* Settings for THIS-PROCEDURE
      Type: Procedure
      Allow: 
      Frames: 0
      Add Fields to: Neither
      Other Settings: CODE-ONLY COMPILE
    */   
   /* *************************  Create Window  ************************** */   
   /* DESIGN Window definition (used by the UIB) 
     CREATE WINDOW Procedure ASSIGN
            HEIGHT             = 18.71
            WIDTH              = 58.4.
   _* END WINDOW DEFINITION *_
                                                                           */   
   /* ************************* Included-Libraries *********************** */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*--------------------------------------------------------------------------
       File        : appsprop.i
       Purpose     : Defines basic properties.
       Syntax      : {src/adm2/appsprop.i}
   
       Description :
   
       Modified    : 01/24/2001
     ------------------------------------------------------------------------*/   
   /*          This .W file was created with the Progress UIB.             */   
   /*----------------------------------------------------------------------*/   
   /* ***************************  Definitions  ************************** */   
   /* Custom instance definition file */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*--------------------------------------------------------------------------
       File        : appserverdefscustom.i
       Purpose     : Entry point to define:
                       New definitions
                       {&xcInstanceProperties}
                       {&ADM-PROPERTY-DLG}
       Syntax      : {src/adm2/custom/appserverdefscustom.i}
   
       Modified    : 01/24/2001
       
       Notes       : Referenced in {src/adm2/appsprop.i}
     ------------------------------------------------------------------------*/   
   /* _UIB-CODE-BLOCK-END */   
   /* ********************  Preprocessor Definitions  ******************** */   
   /* _UIB-PREPROCESSOR-BLOCK-END */   
   /* *********************** Procedure Settings ************************ */   
   /* Settings for THIS-PROCEDURE
      Type: Include
      Allow: 
      Frames: 0
      Add Fields to: Neither
      Other Settings: INCLUDE-ONLY
    */   
   /* *************************  Create Window  ************************** */   
   /* DESIGN Window definition (used by the UIB) 
     CREATE WINDOW Include ASSIGN
            HEIGHT             = 7.86
            WIDTH              = 60.
   _* END WINDOW DEFINITION *_
                                                                           */   
   /* ***************************  Main Block  *************************** */   
   /* Include the file which defines prototypes for all of the super
      procedure's entry points. 
      And skip including the prototypes if we are *any* super procedure. */   
   /* These preprocessors tell at compile time which properties can
        be retrieved directly from the temp-table */   
   /* include properties from smart */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*--------------------------------------------------------------------------
       File        : smrtprop.i
       Purpose     : Starts smart.p super procedure and defines general
                     SmartObject properties and other values.
       Syntax      : {src/adm2/smrtprop.i}
   
       Description :
   
       Modified    : February 4, 2000 - Version 9.1B
     ------------------------------------------------------------------------*/   
   /*          This .W file was created with the Progress UIB.             */   
   /*----------------------------------------------------------------------*/   
   /* ***************************  Definitions  ************************** */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*--------------------------------------------------------------------------
       File        : smartdefscustom.i
       Purpose     : Entry point to define:
                       New definitions
                       {&xcInstanceProperties}
                       {&ADM-PROPERTY-DLG}
       Syntax      : {src/adm2/custom/smartdefscustom.i}
   
       Modified    : 06/03/1999
       
       Notes       : Referenced in {src/adm2/smrtprop.i}
     ------------------------------------------------------------------------*/   
   /* Service Managers */   
   /*********************************************************************
   * Copyright (C) 2000-2002 by Progress Software Corporation ("PSC"),  *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*---------------------------------------------------------------------------------
     File: globals.i
   
     Description:  ICF Global Shared Variables
   
     Purpose:      ICF Global Shared Variables
                   The global shared variables available within the Dynamics environment.
                   
                   NOTE: (Dynamic Only )
                   Shared variables do not span the client / server boundary. On the
                   client these are set-up in afstartupp.p and on the server in the agent
                   as_startup.p
                   Also note if we include the guts of a procedure as an internal
                   procedure for shared client/server code, we sometime need to
                   ensure the shared variables are not redefined.
   
     Parameters:   <none>
   
     History:
     --------
   
     Modified: 11/08/2001 - Mark Davies (MIP)
               Copied code for ICF afglobals.i and added here to allow
               ICF changed adm code to compile. afglobals.i now includes
               this file for definitions.
               Any ADM code will now reference globals.i which excludes any other
               ICF specific includes.
   -----------------------------------------------------------*/   
   /*                   This .W file was created with the Progress UIB.             */   
   /*-------------------------------------------------------------------------------*/   
   /* ***************************  Definitions  ************************** */   
   /* _UIB-CODE-BLOCK-END */   
   /* ********************  Preprocessor Definitions  ******************** */   
   /* _UIB-PREPROCESSOR-BLOCK-END */   
   /* *********************** Procedure Settings ************************ */   
   /* Settings for THIS-PROCEDURE
      Type: Include
      Allow: 
      Frames: 0
      Add Fields to: Neither
      Other Settings: INCLUDE-ONLY
    */   
   /* *************************  Create Window  ************************** */   
   /* DESIGN Window definition (used by the UIB) 
     CREATE WINDOW Include ASSIGN
            HEIGHT             = 5.57
            WIDTH              = 64.4.
   _* END WINDOW DEFINITION *_
                                                                           */   
   /* ***************************  Main Block  *************************** */   
   /* Used whereever Crystal is used and when defining the tt_datasource temp-table 
      to define the maximum columns supported when printing to crystal.
      See browsercustom.p and afcrpplip2p.p for example uses
   */   
   /* Used for filter checks */   
   /* The table delimiter used in assignlist, *ColumnsByTable props */   
   handle gshAstraAppserver = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshastraappserver", true, new handle());

   /* Handle to Astra Application Server Partition */   
   handle gshSessionManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshsessionmanager", true, new handle());

   /* Handle Astra Session Manager */   
   handle gshRimanager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshrimanager", true, new handle());

   /* Handle Astra Session Manager */   
   handle gshSecurityManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshsecuritymanager", true, new handle());

   /* Handle Astra Security Manager */   
   handle gshProfileManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshprofilemanager", true, new handle());

   /* Handle Astra Profile Manager */   
   handle gshRepositoryManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshrepositorymanager", true, new handle());

   /* Handle Astra Repository Manager */   
   handle gshTranslationManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshtranslationmanager", true, new handle());

   /* Handle Astra Translation Manager */   
   handle gshWebManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshwebmanager", true, new handle());

   /* Handle Astra Web Manager */   
   character gscSessionId = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gscsessionid", true, new character(""));

   /* Unique session id */   
   decimal gsdSessionObj = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gsdsessionobj", true, new decimal(0));

   /* Unique session id */   
   handle gshFinManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshfinmanager", true, new handle());

   /* Handle AstraFin Generic Manager */   
   handle gshGenManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshgenmanager", true, new handle());

   /* Handle AstraFramework Generic Manager */   
   handle gshAgnManager = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gshagnmanager", true, new handle());

   /* Handle AstraGen Generic Manager */   
   decimal gsdTempUniqueId = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gsdtempuniqueid", true, new decimal(0));

   /* A number for unique objects */   
   decimal gsdUserObj = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gsduserobj", true, new decimal(0));

   /* User Object ID */   
   decimal gsdRenderTypeObj = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gsdrendertypeobj", true, new decimal(0));

   /* Render Type Object ID */   
   decimal gsdSessionScopeObj = SharedVariableManager.addVariable(ScopeLevel.GLOBAL, "gsdsessionscopeobj", true, new decimal(0));

   /* Render Type Object ID */   
   /*********************************************************************
   * Copyright (C) 2000 by Progress Software Corporation ("PSC"),       *
   * 14 Oak Park, Bedford, MA 01730, and other contributors as listed   *
   * below.  All Rights Reserved.                                       *
   *                                                                    *
   * The Initial Developer of the Original Code is PSC.  The Original   *
   * Code is Progress IDE code released to open source December 1, 2000.*
   *                                                                    *
   * The contents of this file are subject to the Possenet Public       *
   * License Version 1.0 (the "License"); you may not use this file     *
   * except in compliance with the License.  A copy of the License is   *
   * available as of the date of this notice at                         *
   * http://www.possenet.org/license.html                               *
   *                                                                    *
   * Software distributed under the License is distributed on an "AS IS"*
   * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. You*
   * should refer to the License for the specific language governing    *
   * rights and limitations under the License.                          *
   *                                                                    *
   * Contributors:                                                      *
   *                                                                    *
   *********************************************************************/   
   /*------------------------------------------------------------------------
       File        : globalscustom.i
       Purpose     : The developer can create his own version of this to 
                     allow 'global' definitions to be added to the 
                     framework.   
       Syntax      : {src/adm2/custom/globalscustom.i}
   
       Description :
   
       Created     : 06/03/1999
       Notes       : Referenced in {src/adm2/globals.i}
     ----------------------------------------------------------------------*/   
   /*          This .W file was created with the Progress AppBuilder.      */   
   /*----------------------------------------------------------------------*/   
   /* ***************************  Definitions  ************************** */   
   /* _UIB-CODE-BLOCK-END */   
   /* define the ADM Version and broker handle for all SmartObjects */   
   /* Note that adm-props-from-repository is documented as possible to define
        as a variable for flexible override, so it cannot be used for conditional 
        compile.  */   
   /* adm-props-from-repository is supported to be defined as a variable,
       so it is not used for conditional compile */   
   handle ghProp = TypeFactory.handle();

   /**
    * External procedure (converted to Java from the 4GL source code
    * in common/adm2/appserver.p).
    */
   public void execute()
   {
      externalProcedure(Appserver.this, new Block((Init) () -> 
      {
         gsdSessionObj.setPrecision(9);
         gsdTempUniqueId.setPrecision(9);
         gsdUserObj.setPrecision(9);
         gsdRenderTypeObj.setPrecision(9);
         gsdSessionScopeObj.setPrecision(9);
      }, 
      (Body) () -> 
      {
         if (_not(appSrvUtils.isValid()))
         {
            ControlFlowOps.invokePersistentSet("adecomm/as-utils.w", appSrvUtils);
         }
         
         thisProcedure().unwrapCommonProcedure().addSuperProcedure(appSrvUtils);
         
         /* _UIB-CODE-BLOCK-END */
         /* *********************** Procedure Settings ************************ */
         /* Settings for THIS-PROCEDURE
            Type: Method-Library
            Allow: 
            Frames: 0
            Add Fields to: Neither
            Other Settings: INCLUDE-ONLY
          */
         /* *************************  Create Window  ************************** */
         /* DESIGN Window definition (used by the UIB) 
           CREATE WINDOW Method-Library ASSIGN
                  HEIGHT             = 14.86
                  WIDTH              = 57.2.
         _* END WINDOW DEFINITION *_
                                                                                 */
         /* ***************************  Main Block  *************************** */
         /* Include the file which defines prototypes for all of the super
              procedure's entry points. Also, start or attach to the super procedure.
              Skip start-super-proc if we *are* the super procedure. 
              And skip including the prototypes if we are *any* super procedure. */
         /* These preprocessors let the get and set methods know at compile time
             which property values are located in the temp-table and which must
             be accessed through the property functions.
          */
         /* for support of dynamic objects */
         /* physical name - no path */
         /* for support of runtime parameters passed from the menus */
         /* Properties for super procedure management */
         /* The name of the super procedure; can be a CSV list */ 
 /* STATELESS, STATEFUL; the mode that all supers for the object will be launched in */ 
 /* Runtime property storing handles to the object's super procedures, defined in SuperProcedure. */
         /* Layout support for dynamic objects */
         /* contains a layout position that allows an object to be laid out correctly. */
         /* determine if we need to create the prop fields here or from the */
         /* repository (if ICF is running ) */
         /* COMMENT THE FOLLOWING LINE TO USE THE "<class>prop.i" DEFINITIONS */
         /* This temp-table defines all the propertt fields for an object.
             This include file contributes the DEFINE statement header and
             all basic smart object properties. Each other property class include file
             adds its own fields and then the parent object ends the statement.
             Define the fields for smart objects only, not for their super procedures.
          */
         if (_not(handle.fromString(entry(1, thisProcedure().unwrapADMData().getADMData(), chr(1))).isValid()))
         {
         }
         
         /* _UIB-CODE-BLOCK-END */
         if (_not(handle.fromString(entry(1, thisProcedure().unwrapADMData().getADMData(), chr(1))).isValid()))
         {
         }
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * ***********************  Function Implementations****************
    */
   public character getObjectType()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns the Type of the SmartObject, such as "SmartDataObject", etc.
                  For an ADM Super procedure, this function returns "SUPER".
         Params:  <none>
      ------------------------------------------------------------------------------*/
      character cType = TypeFactory.character();
      
      return function("getObjectType", character.class, new Block((Body) () -> 
      {
         cType.assign("SUPER");
         returnNormal(cType);
      }));
   }

   /** *********************  Internal Procedures ********************** */
   /**
    * Purpose:  Bind the object to the server.  <br>
    * Parameters:  &lt;none&gt;<br>
    * Notes:     This provides the same functionality as getAsHandle, but does not<br>
    *            expose the handle and is considered to be the API for an outside <br>
    *            caller.
    */
   public void bindServer()
   {
      character cBindScope = TypeFactory.character();
      integer iStack = TypeFactory.integer();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* {fn <func-name> [handle] */ 
         ControlFlowOps.invokeDynamicFunctionIn(handle.class, "getASHandle", targetProcedure());
         
         if (_not(new logical( /* {fn <func-name> [handle] */ ControlFlowOps.invokeDynamicFunctionIn(logical.class, "anyMessage", targetProcedure()))))
         {
            /* {get Prop var [handle]} */ 
            cBindScope.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getBindScope", targetProcedure()));
            
            /* if Bindscope is not blank then the signature should already indicate 
                    where the unbind should take place 
                    Loop to ignore possible overrides of bindServer
                    (10 is just a lazy and paranoid do while true ) */
            if (_isEqual(cBindScope, ""))
            {
               loopLabel0:
               for (iStack.assign(2); _isLessThanOrEqual(iStack, 10); iStack.increment())
               {
                  if (_not(begins(EnvironmentOps.getSourceName(iStack), "bindServer")))
                  {
                     ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                     ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
                     ghProp.unwrapBufferField().changeValue(EnvironmentOps.getSourceName(iStack));
                     break loopLabel0;
                  }
               }
            }
         }
         else
         {
            returnError("ADM-ERROR");
         }
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * Purpose:    connect and get the handle of the appServer <br>
    * Parameters:  &lt;none&gt;<br>
    * Notes:      If AsDivison is blank, prompt parameters will be passed to<br>
    *             the global appserverConnect. <br>
    *             (Note that these parameters are ignored in ICF and are <br>
    *              only kept for backwards compatibility)  <br>
    *             This is part of getAsHandle, which returns ? to indicate errors<br>
    *             so most errors are just returned.
    */
   public void connectServer(final handle phAppService_1)
   {
      logical lAsusePrompt = TypeFactory.logical();
      character cAsinfo = TypeFactory.character();
      character cAppService = TypeFactory.character();
      handle hContainer = TypeFactory.handle();
      logical lAsHasConnected = TypeFactory.logical();
      
      internalProcedure(new Block((Init) () -> 
      {
         TypeFactory.initOutput(phAppService_1);
      }, 
      (Body) () -> 
      {
         /* Silent error. See notes */
         /* {get Prop var [handle]} */ 
         cAppService.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getAppService", targetProcedure()));
         
         /* We allow to borrow the container's appservice and do a separate connection
                independently of the container if someone would want to do that.  */
         if (_isEqual(cAppService, ""))
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ContainerSource"));
            hContainer.assign(ghProp.unwrapBufferField().value());
            /* {get Prop var [handle]} */
            silent(() -> cAppService.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getAppService", hContainer)));
         }
         
         if (_or(isEqual(cAppService, ""), () -> isUnknown(cAppService)))
         {
            returnNormal();
         }
         
         /* {get Prop var [handle]} */ 
         lAsHasConnected.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getAsHasConnected", targetProcedure()));
         
         if (_not(lAsHasConnected))
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ASUsePrompt"));
            lAsusePrompt.assign(ghProp.unwrapBufferField().value());
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ASInfo"));
            cAsinfo.assign(ghProp.unwrapBufferField().value());
            lAsusePrompt.assign(_isEqual(lAsusePrompt, false) ? new logical() : lAsusePrompt);
            cAsinfo.assign(_isEqual(cAsinfo, "") ? new character() : cAsinfo);
         }
         else
         {
            lAsusePrompt.setUnknown();
            cAsinfo.setUnknown();
         }
         
         ControlFlowOps.invokeWithMode("appServerConnect", "IIIO", cAppService, lAsusePrompt, cAsinfo, phAppService_1);
         
         if (_isEqual(ControlFlowOps.getReturnValue(), "ERROR"))
         {
            ControlFlowOps.invokeInWithMode("addMessage", targetProcedure(), "III", concat(cAppService, " partition is running locally without the proper database connection(s)."), new unknown(), new unknown());
            returnError("ADM-ERROR");
         }
 /* or phAppService = SESSION:HANDLE*/;
         
         /* hAppService = session or return-value = 'error' */
         if (_not(lAsHasConnected))
         {
            if (_and(notUnknown(phAppService_1), () -> isNotEqual(phAppService_1, SessionUtils.asHandle())))
            {
               /* {set Prop value [handle]} */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setASDivision", targetProcedure(), "I", "Client");
            }
            else
            {
               /* {set Prop value [handle]} */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setAsDivision", targetProcedure(), "I", "");
            }
         }
         returnNormal();
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * Purpose:     Disconnects the AppServer connection if present before invoking <br>
    *              the standard destroy code.<br>
    * <p>
    * Parameters:  &lt;none&gt;
    */
   public void destroyObject()
   {
      internalProcedure(new Block((Body) () -> 
      {
         ControlFlowOps.invokeIn("disconnectObject", targetProcedure());
         /* Disc. from AppServer */
         silent(() -> ControlFlowOps.runSuper("destroyObject"));
         
         if (_or(ErrorManager.isError(), () -> isEqual(ControlFlowOps.getReturnValue(), "ADM-ERROR")))
         {
            returnError("ADM-ERROR");
         }
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * Purpose:  Destroy the server object and retrieve its context.    <br>
    * Parameters:  &lt;none&gt;<br>
    * Notes:    Called from unbindServer when stateless.  <br>
    *         - The intention is to move context management up to this class<br>
    *           from data.p and sbo.p, but as of current this method calls <br>
    *           APIs that is not implemented here, but only exists in QueryObjects <br>
    *           if QueryObject is true  <br>
    *         - This event is not always called as queryobjects now supports <br>
    *           'one-hit' datarequests or call destroyObject directly <br>
    *           when BindScope is 'data'.
    */
   public void destroyServerObject()
   {
      logical lAsbound = TypeFactory.logical();
      handle hAshandle = TypeFactory.handle();
      character cContext = TypeFactory.character();
      logical lNeedContext = TypeFactory.logical();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* {get Prop var [handle]} */ 
         lAsbound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getASBound", targetProcedure()));
         
         if ((lAsbound).booleanValue())
         {
            /* {get Prop var [handle]} */ 
            hAshandle.assign(ControlFlowOps.invokeDynamicFunctionIn(handle.class, "getASHandle", targetProcedure()));
            /* {get Prop var [handle]} */ 
            lNeedContext.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getNeedContext", targetProcedure()));
            
            /* QueryObjects like the sdo and sbo 
                                                   need context, so we assume context 
                                                   methods are available even if not yet 
                                                   implemented in this class */
            if ((lNeedContext).booleanValue())
            {
               ControlFlowOps.invokeInWithMode("getContextAndDestroy", hAshandle, "O", cContext);
               /* {fnarg <func-name> <arg> [handle] */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "applyContextFromServer", targetProcedure(), "I", cContext);
            }
            else
            {
               ControlFlowOps.invokeIn("destroyObject", hAshandle);
            }
            
            /* getAsHandle checks valid-handle, but handles are resused so make sure
                   that valid-handle returns false */
            /* {set Prop value [handle]} */ 
            ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setAsHandle", targetProcedure(), "I", new unknown());
         }
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * Purpose:     To disconnect from the AppServer if there is one. <br>
    * Parameters:  &lt;none&gt;<br>
    * Notes:       If lAsbound we do an explicit destroyObject on the AppServer <br>
    *              to give that object an opportunity to clean up.                <br>
    *              This procedure is invoked from destroyObject, but can also be <br>
    *              run directly to disconnect without exiting.
    */
   public void disconnectObject()
   {
      handle hAshandle = TypeFactory.handle();
      logical lAsbound = TypeFactory.logical();
      character cAppService = TypeFactory.character();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* we might be connected even if we have no Appservice 
             (borrowed from container) */
         /* {get Prop var [handle]} */ 
         lAsbound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getASBound", targetProcedure()));
         
         if ((lAsbound).booleanValue())
         {
            /* {get Prop var [handle]} */ 
            hAshandle.assign(ControlFlowOps.invokeDynamicFunctionIn(handle.class, "getASHandle", targetProcedure()));
            
            if (_and(hAshandle.isValid(), () -> isNotEqual(hAshandle, targetProcedure())))
            {
               ControlFlowOps.invokeIn("destroyObject", hAshandle);
               /* ensure AsHandle is not-valid  */
               /* {set Prop value [handle]} */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setAsHandle", targetProcedure(), "I", new unknown());
            }
         }
         /* if lAsBound  */
         /* {get Prop var [handle]} */ 
         cAppService.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getAppService", targetProcedure()));
         
         if (_isNotEqual(cAppService, ""))
         {
            ControlFlowOps.invokeWithMode("appServerDisconnect", "I", cAppService);
         }
         
         returnNormal();
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    *   Purpose: Common 'event' called after any data request. Both read and save of <br>
    *            data are considered to be data requests.                    <br>
    *     Notes: The task is to ensure that a stateless connection is unbound correctly <br>
    *            with a call to the unbindServer procedure, which has the logic to <br>
    *            unbind or not.  <br>
    *          - Even if this event implies that the data request is over, the <br>
    *            BindScope property must be set to 'Data' to ensure an immediate <br>
    *            unbind by passing 'UNCONDITIONAL' to unbindServer. <br>
    *          - The adm sets BindScope to 'data' as early as possible in all methods <br>
    *            that eventually will cause a data request.<br>
    *          - The appserver class has no data requests, but the logic here is<br>
    *            common for both data.p and sbo.p <br>
    *          - Historically, this event's main purpose was to retrieve context from <br>
    *            the server if still bound, to ensure that important properties were <br>
    *            returned as early as possible, but this is no longer required as the<br>
    *            new server data requests implemented for 9.1D all have context as <br>
    *            input-output parameter. <br>
    * Note Date: 2002/02/05
    */
   public void endClientDataRequest()
   {
      character cAsDivision = TypeFactory.character();
      character cOperatingMode = TypeFactory.character();
      character cBindScope = TypeFactory.character();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* {get Prop var [handle]} */ 
         cAsDivision.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getAsDivision", targetProcedure()));
         
         /* Get out if not 'client' */
         if (_isNotEqual(cAsDivision, "Client"))
         {
            returnNormal();
         }
         
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         cOperatingMode.assign(ghProp.unwrapBufferField().value());
         
         if (_isEqual(cOperatingMode, "stateless"))
         {
            /* {get Prop var [handle]} */ 
            cBindScope.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getBindScope", targetProcedure()));
            /* We pass the caller to unbindServer to ensure proper unbinding unless
                   the BindScope is 'data', which we use as an indication to pass 
                   'unconditional' to ensure that the unbind takes place independent 
                   of BindSignature. */
            ControlFlowOps.invokeInWithMode("unbindServer", targetProcedure(), "I", _isEqual(cBindScope, "Data") ? new character("UNCONDITIONAL") : EnvironmentOps.getSourceName(2));
         }
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    *    Purpose: Initialize the server object after it has been started/restarted    <br>
    * Parameters: &lt;none&gt;<br>
    *      Notes: An extra call to the Appserver is required on the first call to <br>
    *             retrieve the ServerOperatingMode. This is expensive and could have <br>
    *             been avoided by implementing synchronizeProperties in this class <br>
    *             that returns this property, but most implementors of this will<br>
    *             class either avoid this procedure completely by setting <br>
    *             ASInitializeOnRun to false or override it for more complete <br>
    *             context management so this has not been implemented.<br>
    *           - From 9.1D this is not called as part of a normal data request.      <br>
    *           - This is an internal 'event' that is called from runServerObject <br>
    *             after a successful run on server. It assumes that the server is <br>
    *             bound and silently ignores any calls when not bound.                  <br>
    *           - It is expected that context management will be moved up to this <br>
    *             class (from data.p and sbo.p) sometime in the future, but as of <br>
    *             current this method calls APIs that only exists in 'query objects' <br>
    *             if NeedContext is true. The NeedContext defaults to true if <br>
    *             QueryObject is true.<br>
    * Note date:  2002/02/05
    */
   public void initializeServerObject()
   {
      logical lHasStarted = TypeFactory.logical();
      handle hAshandle = TypeFactory.handle();
      character cOperatingMode = TypeFactory.character();
      logical lAsbound = TypeFactory.logical();
      logical lNeedContext = TypeFactory.logical();
      character cContext = TypeFactory.character();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* {get Prop var [handle]} */ 
         lAsbound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getASBound", targetProcedure()));
         
         /* This call only makes sense when bound to the server */
         if ((lAsbound).booleanValue())
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHasStarted"));
            lHasStarted.assign(ghProp.unwrapBufferField().value());
            /* {get Prop var [handle]} */ 
            hAshandle.assign(ControlFlowOps.invokeDynamicFunctionIn(handle.class, "getASHandle", targetProcedure()));
            /* {get Prop var [handle]} */ 
            lNeedContext.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getNeedContext", targetProcedure()));
            
            /* QueryObjects like the sdo and sbo 
                                                   need context, so we assume context 
                                                   methods are available even if not yet 
                                                   implemented in this class */
            if ((lNeedContext).booleanValue())
            {
               cContext /* {fn <func-name> [handle] */ .assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "obtainContextForServer", targetProcedure()));
               
               /* synchronizeProperties does not currently exist in SBO, so the SBO 
                        overrides this and calls super only if hasStarted  */
               if (_not(lHasStarted))
               {
                  ControlFlowOps.invokeInWithMode("synchronizeProperties", hAshandle, "IO", cContext, cContext);
                  /* {fnarg <func-name> <arg> [handle] */ 
                  ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "applyContextFromServer", targetProcedure(), "I", cContext);
               }
               else
               {
                  ControlFlowOps.invokeInWithMode("setContextAndInitialize", hAshandle, "I", cContext);
               }
            }
            else
            {
               ControlFlowOps.invokeIn("initializeObject", hAshandle);
               
               if (_not(lHasStarted))
               {
                  cOperatingMode.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getServerOperatingMode", hAshandle));
                  ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                  ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
                  ghProp.unwrapBufferField().changeValue(cOperatingMode);
               }
            }
         }
         returnNormal();
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * Purpose:    When a SmartDataObject is split and running Statelessly on an<br>
    *             AppServer, it is shutdown after each use and then restarted for<br>
    *             the next. restartServerObject is run on the client to restart<br>
    *             the SmartDataObject on the server.<br>
    * Parameters:  <br>
    * Notes:      The startServerObject will (should) be run on the first connection. <br>
    *           - Both start- and restart- runs runServerObject, which does the <br>
    *             actual run and the optional initialize.    <br>
    *           - This procedure is considered to be an 'internal event' and <br>
    *             does silently return if the object is bound or Appservice = ''<br>
    *             and does not check if ASHasStarted is true (As this may not be <br>
    *             damaging if there's no difference between start/restart). <br>
    *             It is assumed to be called only when unbound and ASHasStarted is <br>
    *             true as when called from getAsHandle. <br>
    *           - getASHandle is depending on the 'silent error' when Appservice <br>
    *             is blank for backwards compatibility with data and sbo classes, <br>
    *            (not completely true, it actually checked ASDivision='client', but<br>
    *             that is too restrictive) so if errors are added here, make sure <br>
    *             getAsHandle deals with (removes) them before returning
    */
   public void restartServerObject()
   {
      handle hAppService = TypeFactory.handle();
      logical lBound = TypeFactory.logical();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* {get Prop var [handle]} */ 
         lBound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getASBound", targetProcedure()));
         
         if ((lBound).booleanValue())
         {
            returnNormal();
         }
         
         ControlFlowOps.invokeInWithMode("connectServer", targetProcedure(), "O", hAppService);
         
         if (_isUnknown(hAppService))
         {
            returnError("ADM-ERROR");
         }
         
         if ((hAppService.isValid()).booleanValue())
         {
            silent(() -> ControlFlowOps.invokeInWithMode("runServerObject", targetProcedure(), "I", hAppService));
            
            if (_or(ErrorManager.isError(), () -> isEqual(ControlFlowOps.getReturnValue(), "ADM-ERROR")))
            {
               returnError(ControlFlowOps.getReturnValue());
            }
         }
         else
         {
            ControlFlowOps.invokeInWithMode("addMessage", targetProcedure(), "III", concat("Lost connection to partition ", ControlFlowOps.invokeDynamicFunctionIn(character.class, "getAppService", targetProcedure())) /* {fn <func-name> [handle] */, new unknown(), new unknown());
            returnError("ADM-ERROR");
         }
         returnNormal();
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * Purpose:    Run the server part of this object and set AShandle.<br>
    * Parameters: phAppservice - AppServer Session handle<br>
    * Notes:      Called from startServerObject and restartServerObject.
    */
   public void runServerObject(final handle _phAppService_1)
   {
      handle phAppService_1 = TypeFactory.initInput(_phAppService_1);
      character cServerFileName = TypeFactory.character();
      handle hAshandle = TypeFactory.handle();
      logical lInitialize = TypeFactory.logical();
      
      internalProcedure(new Block((Body) () -> 
      {
         if (_and(notUnknown(phAppService_1), () -> isNotEqual(phAppService_1, SessionUtils.asHandle())))
         {
            /* This property (9.1B) holds the file name to run on the AppServer. 
                   Dynamic object names won't be usable (it will be dyn*). */
            /* {get Prop var [handle]} */ 
            cServerFileName.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getServerFileName", targetProcedure()));
            hAshandle.assign(ControlFlowOps.invokeDynamicFunctionInWithMode(handle.class, "runServerProcedure", targetProcedure(), "II", cServerFileName, phAppService_1));
            
            if ((hAshandle.isValid()).booleanValue())
            {
               /* Save the handle of the connected ServerObject */
               /* {set Prop value [handle]} */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setASHandle", targetProcedure(), "I", hAshandle);
               ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
               ghProp.assign(ghProp.unwrapBuffer().bufferField("ASInitializeOnRun"));
               lInitialize.assign(ghProp.unwrapBufferField().value());
               
               if ((lInitialize).booleanValue())
               {
                  ControlFlowOps.invokeIn("initializeServerObject", targetProcedure());
               }
               
               ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
               ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHasStarted"));
               ghProp.unwrapBufferField().changeValue(new logical(true));
            }
            else
            {
               ControlFlowOps.invokeInWithMode("addMessage", targetProcedure(), "III", concat("AppServer startup failed for ", cServerFileName), new unknown(), new unknown());
               returnError("ADM-ERROR");
            }
         }
         else
         {
             /* {set Prop value [handle]} */ ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setASHandle", targetProcedure(), "I", targetProcedure());
         }
         returnNormal();
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public void startServerObject()
   {
      character cOperatingMode = TypeFactory.character();
      handle hAppService = TypeFactory.handle();
      logical lBound = TypeFactory.logical();
      character cBindScope = TypeFactory.character();
      
      internalProcedure(new Block((Body) () -> 
      {
         /* Silent error. See notes */
         /* {get Prop var [handle]} */ 
         lBound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getASBound", targetProcedure()));
         
         if ((lBound).booleanValue())
         {
            returnNormal();
         }
         
         ControlFlowOps.invokeInWithMode("connectServer", targetProcedure(), "O", hAppService);
         
         if (_isUnknown(hAppService))
         {
            returnError("ADM-ERROR");
         }
         
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         cOperatingMode.assign(ghProp.unwrapBufferField().value());
         /* We run this also if not valid handle just to set ASHandle. */
         silent(() -> ControlFlowOps.invokeInWithMode("runServerObject", targetProcedure(), "I", hAppService));
         
         if (_or(ErrorManager.isError(), () -> isEqual(ControlFlowOps.getReturnValue(), "ADM-ERROR")))
         {
            returnError(ControlFlowOps.getReturnValue());
         }
         
         /* Override the actual state if the instance property is to force stateful */
         if (_isEqual(cOperatingMode, "STATE-RESET"))
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
            ghProp.unwrapBufferField().changeValue(cOperatingMode);
         }
         /* {set Prop value [handle]} */ 
         
         /* This procedure may be called from the outside just to initialize 
              properties, ensure that it unbinds if called with no bindScope 
             (Example:
               get* properties in SDOs calls this in the containing SBO in order 
               to get ALL data from the server. This would normally be during 
               initialization in which case BindScope is defined, but it is possible 
               to call them separately) */
         /* {get Prop var [handle]} */ 
         cBindScope.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getBindScope", targetProcedure()));
         
         if (_isEqual(cBindScope, ""))
         {
            ControlFlowOps.invokeInWithMode("unbindserver", targetProcedure(), "I", new unknown());
         }
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    *    Purpose:   To synchronize certain properties on the server side with its<br>
    *               client side.<br>
    * Parameters:<br>
    *     INPUT pcPropertiesForServer  CHAR<br>
    *              - The format is defined by the classes that extends this <br>
    *     OUTPUT pcPropertiesForClient CHAR<br>
    *              - The format is defined by the classes that extends this   <br>
    *              <br>
    *     Notes:   This procedure is designed to be invoked from the client side, but <br>
    *              executed on the server half of a SmartDataObject. While executing <br>
    *              on the server side it runs setContextAndInitialize to set the <br>
    *              passed properties, then calls obtainContextForClient to accumulate <br>
    *              properties to return in the output parameter to be applied on the <br>
    *              client.  <br>
    *          -   See obtainContextForClient for further details on when and how <br>
    *              special 'first request' data is returned.<br>
    *          -   This is one of many variations on client server communication and<br>
    *              is now only used on the first request and even in that case only <br>
    *              if getAsHandle is being called prior to the actual data request.  <br>
    *          -   This is using APIs that (not yet) exists in this clas, but only<br>
    *              in classes that NeedContext (containr, data  and sbo)                <br>
    * Note date: 2002/02/04
    */
   public void synchronizeProperties(final character _pcPropertiesForServer, final character pcPropertiesForClient)
   {
      character pcPropertiesForServer = TypeFactory.initInput(_pcPropertiesForServer);
      
      internalProcedure(new Block((Init) () -> 
      {
         TypeFactory.initOutput(pcPropertiesForClient);
      }, 
      (Body) () -> 
      {
         ControlFlowOps.invokeInWithMode("setContextAndInitialize", targetProcedure(), "I", pcPropertiesForServer);
         pcPropertiesForClient /* {fn <func-name> [handle] */ .assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "obtainContextForClient", targetProcedure()));
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    *   Purpose:  unbind the AppServer by destroying the server side object <br>
    *             started by the client.        <br>
    * Parameters: pcMode - Rules for unbinding<br>
    *                     'conditional' or  ?   <br>
    *                       - Unbind if called from same level as BindSignature<br>
    *                         unless BindScope is 'STRONG'. <br>
    *                     'unconditional' <br>
    *                       - Unbind unconditionally                       <br>
    *                      program-name(x)) or any space delimited string <br>
    *                       - Unbind if passed string  =  BindSignature<br>
    *                         Only if BindScope is blank.<br>
    * Notes:     This procedure allows nested calls of procedures that binds and <br>
    *            unbinds, but still postpone the unbinding until we are back <br>
    *            at the level that did the actual binding.<br>
    *         -  The logic in this procedure is dependent of the fact that  <br>
    *            getAsHandle or bindServer does the actual binding (if AsHandle is ?)<br>
    *            and logs the call level by setBindSignature = program-name(2).<br>
    *            or that setBindScope defines the call level by <br>
    *            setBindSignature = program-name(2). <br>
    *         -  An external caller will use the following sequence to ensure that <br>
    *            all calls are done with one connection. <br>
    *             <br>
    *             RUN bindServer in &lt;handle&gt;.<br>
    *             &lt;some request&gt; in &lt;handle&gt;<br>
    *             &lt;some request&gt; in &lt;handle&gt;<br>
    *             RUN unbindServer(?). <br>
    *            <br>
    *            <br>
    *         -  Internal calls will typically look like this.<br>
    *            <br>
    *            hAsHandle = getAsHandle(). <br>
    *            &lt;some request&gt; in hasHandle <br>
    *            RUN unbindServer(?).  <br>
    *                                                  <br>
    *         - Currently overrides need to do the following <br>
    *             <br>
    *            def input param pcmode as char.<br>
    *             run SUPER(if pcMode = ? then program-name(2) else pcMode). <br>
    *             <br>
    *         - Limitations applies as recursive calls and external callers have<br>
    *           the same signature, so an unbind may happen to early. <br>
    *          (This can probably be fixed by saving the complete stack in<br>
    *           BindSignature)
    */
   public void unbindServer(final character _pcMode)
   {
      character pcMode = TypeFactory.initInput(_pcMode);
      character cAsDivision = TypeFactory.character();
      character cServerOperatingMode = TypeFactory.character();
      handle hAsHandle = TypeFactory.handle();
      character cBindSignature = TypeFactory.character();
      character cBindScope = TypeFactory.character();
      logical lAsbound = TypeFactory.logical();
      character cCaller = TypeFactory.character();
      integer iStack = TypeFactory.integer();
      
      internalProcedure(new Block((Body) () -> 
      {
         if (_isEqual(pcMode, ""))
         {
            /* We reserve this for the future in the case we need an alternative easy 
                   to use parameter.  */
            /* {fnarg <func-name> <arg> [handle] */ 
            messageBox(substitute(ControlFlowOps.invokeDynamicFunctionInWithMode(character.class, "messageNumber", targetProcedure(), "I", 41), EnvironmentOps.getSourceName(2)), ALERT_WARNING, BTN_OK, (String) null);
            returnNormal();
         }
         
         /****** Future ? 
           We could fix this so that overrides was overlooked;
           something like this + a check of the targets super-procedures...    
           i = 2.
           callerIp = entry(1,PROGRAM-NAME(2),'').
           cProgramName = PROGRAM-NAME(i).
           DO WHILE program-name(1) BEGINS callerIp:
             i = i + 1.  
             callerIp = ENTRY(1,PROGRAM-NAME(i)," ":U).
             cProgramName = PROGRAM-NAME(i).
           END.
           ***********/
         /* {get Prop var [handle]} */ 
         cAsDivision.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getASDivision", targetProcedure()));
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         cServerOperatingMode.assign(ghProp.unwrapBufferField().value());
         
         /* Shutdown Server side SDO if stateless client is bound */
         if (_and(isEqual(lookup(cServerOperatingMode, "state-aware,state-reset"), 0), () -> isEqual(cAsDivision, "CLIENT")))
         {
            loopLabel1:
            for (iStack.assign(2); _isLessThanOrEqual(iStack, 10); iStack.increment())
            {
               if (_not(begins(EnvironmentOps.getSourceName(iStack), "unbindServer")))
               {
                  cCaller.assign(EnvironmentOps.getSourceName(iStack));
                  break loopLabel1;
               }
            }
            
            /* {get Prop var [handle]} */ 
            cBindScope.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getBindScope", targetProcedure()));
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
            cBindSignature.assign(ghProp.unwrapBufferField().value());
            
            /* Destroy server side procedure if Mode 'unconditional' or if this is 
                     called from the level that did the binding unless the scope is STRONG*/
            /* This is intended to unbind with a passed signature, 
                           but ALSO when bindSignature is ? and pcMode is ? */
            /* This is intended to unbind with a passed signature, 
                           but ALSO when bindSignature is ? and pcMode is ? */
            if (_or(begins(pcMode, "unconditional"), () -> and(isNotEqual(cBindScope, "STRONG"), () -> or(isEqual(pcMode, cBindSignature), () -> isEqual(cBindSignature, cCaller)))))
            {
               /* {get Prop var [handle]} */ 
               lAsbound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getASBound", targetProcedure()));
               
               if ((lAsbound).booleanValue())
               {
                  /* We assume that we already have context if unbinding a data scope*/
                  if (_isEqual(cBindScope, "Data"))
                  {
                     /* {get Prop var [handle]} */ 
                     hAsHandle.assign(ControlFlowOps.invokeDynamicFunctionIn(handle.class, "getASHandle", targetProcedure()));
                     ControlFlowOps.invokeIn("destroyObject", hAsHandle);
                  }
                  else
                  {
 /* Call the get context and destroy  */
                     ControlFlowOps.invokeIn("destroyServerObject", targetProcedure());
                  }
               }
               /* We reset the BindScope even if we are not bound if we are at the 
                        defined unbind level */
               /* {set Prop value [handle]} */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setBindScope", targetProcedure(), "I", "");
            }
         }
         returnNormal();
      }));
   }

   /** _UIB-CODE-BLOCK-END */
   /**
    * ***********************  Function Implementations****************
    */
   public character getAppService()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns the property naming the logical partition name of
                  the AppService to be used to connect to an AppServer
         Params:  none
                  No xp preprocessor in order not to return '(None)' as saved from 
                  the Instance Property dialog. 
      ------------------------------------------------------------------------------*/
      character cAppService = TypeFactory.character();
      
      return function("getAppService", character.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("AppService"));
         cAppService.assign(ghProp.unwrapBufferField().value());
         /* Probably not required to check for (none) here as we fix this in set,
              and Inst Props from containers uses set, but just in case someone actually 
              stores Inst Props in a db/repository..*/
         returnNormal(_isNotEqual(cAppService, "(None)") /* Untranslatable in the dialog also */ ? cAppService : new character(""));
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getAsbound()
   {
      /*------------------------------------------------------------------------------
        Purpose: Returns TRUE if this object binds the AppServer with a persistently
                 running procedure, usually the 'server' part of this object.              
          Notes: Uses the AsHandle property to check this. 
      ------------------------------------------------------------------------------*/
      handle hAsHandle = TypeFactory.handle();
      
      return function("getASBound", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHandle"));
         hAsHandle.assign(ghProp.unwrapBufferField().value());
         returnNormal(and(hAsHandle.isValid(), () -> isNotEqual(hAsHandle, targetProcedure())));
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public character getAsDivision()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns a string indicating which side of the AppServer this
                  Object is running on; 'Client', 'Server', or none.
         Params:  <none>
           Note:  This is the main condition used in code to separate/trigger 
                  client or server specific logic.
                  It's immediately set to 'server' at start up if session:remote. 
                  On the client, we traditionally connected in initializeObject 
                  and immediately set this property there, but as we now postpone the
                  connection, this value is now unknown (default) until we know a 
                  connection has succeeded. In order for client specific logic to 
                  function as before also when not yet connected, we return 'client' 
                  when the property is unknown. (otherwise we would never connect...)            
                  It's set to blank immediately if Appservice is blank, but otherwise 
                  it is kept as unknown until we REALLY know.
                - getAsHasConnected uses this unknown value to return false until we 
                  have connected!   
      ------------------------------------------------------------------------------*/
      character cDivision = TypeFactory.character();
      character cAppService = TypeFactory.character();
      logical lAsHasStarted = TypeFactory.logical();
      logical lInitialized = TypeFactory.logical();
      
      return function("getAsDivision", character.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASDivision"));
         cDivision.assign(ghProp.unwrapBufferField().value());
         
         if (_isUnknown(cDivision))
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ObjectInitialized"));
            lInitialized.assign(ghProp.unwrapBufferField().value());
            
            if (_not(lInitialized))
            {
               returnNormal(cDivision);
            }
            
            /* If an appservice is defined and we have not started the value is still
                   unknown, but we must assume that we are on a client as this is the 
                   main conditional check that triggers a connection */
            /* {get Prop var [handle]} */ 
            cAppService.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getAppService", targetProcedure()));
            
            if (_isNotEqual(cAppService, ""))
            {
               ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
               ghProp.assign(ghProp.unwrapBuffer().bufferField("AsHasStarted"));
               lAsHasStarted.assign(ghProp.unwrapBufferField().value());
               
               if (_and(not(lAsHasStarted), () -> isEqual( /* {fn <func-name> [handle] */ ControlFlowOps.invokeDynamicFunctionIn(character.class, "getUIBMode", targetProcedure()), "")))
               {
                  cDivision.assign("Client");
               }
               else
               {
                                    /* Something has gone astray if we get here, but whatever reason
                                we certainly do not have a 'client', so return blank, but keep
                                the property value as is, since we really have no clue  */cDivision.assign("");
               }
            }
            else
            {
               cDivision.assign("");
               /* {set Prop value [handle]} */ 
               ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setAsDivision", targetProcedure(), "I", cDivision);
            }
         }
         /* cDivision = ? */
         returnNormal(cDivision);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public handle getAshandle()
   {
      /*------------------------------------------------------------------------------
        Purpose:     Returns the handle to this object's companion procedure 
                     (the copy of itself) running on the AppServer.
       
        Parameters:  <none>
        
        Notes:       Note that what is returned is *not* the handle of the connection,
                     but rather the persistent procedure handle in which internal 
                     procedures and functions can be invoked.           
      ------------------------------------------------------------------------------*/
      handle hAs = TypeFactory.handle();
      logical lHasStarted = TypeFactory.logical();
      character cBindScope = TypeFactory.character();
      integer iStack = TypeFactory.integer();
      
      return function("getASHandle", handle.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("AsHandle"));
         hAs.assign(ghProp.unwrapBufferField().value());
         
         if (_not(hAs.isValid()))
         {
            /* Perhaps it needs to be re-established */
            /* Store the caller info that unbindServer uses to unbind conditionally 
                    unless the Bindscope is not blank as the signature then already is set
                    and must override this */
            /* {get Prop var [handle]} */ 
            cBindScope.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getBindScope", targetProcedure()));
            
            if (_isEqual(cBindScope, ""))
            {
               loopLabel2:
               for (iStack.assign(2); _isLessThanOrEqual(iStack, 10); iStack.increment())
               {
                  if (_not(begins(EnvironmentOps.getSourceName(iStack), "getAsHandle")))
                  {
                     ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                     ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
                     ghProp.unwrapBufferField().changeValue(EnvironmentOps.getSourceName(iStack));
                     break loopLabel2;
                  }
               }
            }
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHasStarted"));
            lHasStarted.assign(ghProp.unwrapBufferField().value());
            
            /* Both start and restart will call connectServer and runServerObject */
            if ((lHasStarted).booleanValue())
            {
               ControlFlowOps.invokeIn("restartServerObject", targetProcedure());
            }
            else
            {
               ControlFlowOps.invokeIn("startServerObject", targetProcedure());
            }
            
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("AsHandle"));
            hAs.assign(ghProp.unwrapBufferField().value());
         }
         /* If it needs to be re-established */
         returnNormal(hAs);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getAsHasConnected()
   {
      /*------------------------------------------------------------------------------
        Purpose: Return yes if a connection has taken place 
          Notes: READ-ONLY property.  
                 Does only know that a connection HAS taken place. Not if we still
                 are connected.
                 This is resolved through the INTERNAL value of AsDivision as this 
                 is unknown until connectServer has connected     
      ------------------------------------------------------------------------------*/
      character cDivision = TypeFactory.character();
      
      return function("getAsHasConnected", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("AsDivision"));
         cDivision.assign(ghProp.unwrapBufferField().value());
         returnNormal(notUnknown(cDivision));
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getAshasStarted()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns (LOGICAL) a flag indicating whether the object has
                  done its first call to its server side object.
         Params:  <none>
          Notes:          
      ------------------------------------------------------------------------------*/
      logical lHasStarted = TypeFactory.logical();
      
      return function("getASHasStarted", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHasStarted"));
         lHasStarted.assign(ghProp.unwrapBufferField().value());
         returnNormal(lHasStarted);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public character getAsinfo()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns the value of the AppServer Information string, if any,
                  used as a parameter when this object connects to its 
                  Partition for the first time.
         Params:  <none>
      ------------------------------------------------------------------------------*/
      character cAsinfo = TypeFactory.character();
      
      return function("getASInfo", character.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASInfo"));
         cAsinfo.assign(ghProp.unwrapBufferField().value());
         returnNormal(cAsinfo);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getAsinitializeOnRun()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns the flag on or off which indicates whether runServerObject             
                  should call initializeServerObject
         Params:  <none>
           Note:  initializeServerObject is called on the client, but will usually 
                  have a call to the server to set and/or retrieve context.  
      ------------------------------------------------------------------------------*/
      logical lInitialize = TypeFactory.logical();
      
      return function("getASInitializeOnRun", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASInitializeOnRun"));
         lInitialize.assign(ghProp.unwrapBufferField().value());
         returnNormal(lInitialize);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getAsusePrompt()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns (LOGICAL) a flag indicating whether the supporting code
                  should prompt for a Username and Password, when connecting to
                  its Application Partition.
         Params:  <none>
      ------------------------------------------------------------------------------*/
      logical lAsuse = TypeFactory.logical();
      
      return function("getASUsePrompt", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASUsePrompt"));
         lAsuse.assign(ghProp.unwrapBufferField().value());
         returnNormal(lAsuse);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public character getBindScope()
   {
      /*------------------------------------------------------------------------------
        Purpose: Returns the bind scope    
          Notes: This basically decides when a stateless connection should be unbound.
                 See setBindScope  for details   
      ------------------------------------------------------------------------------*/
      character cBindScope = TypeFactory.character();
      
      return function("getBindScope", character.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("BindScope"));
         cBindScope.assign(ghProp.unwrapBufferField().value());
         returnNormal(cBindScope);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getNeedContext()
   {
      /*------------------------------------------------------------------------------
        Purpose: Returns true if the object requires context to be passed back 
                 and forth between client and server. 
          Notes: Will be set to true on first call if not explicitly set (unknown) 
                 and the object is a QueryObject or an appserver aware Container.    
      ------------------------------------------------------------------------------*/
      logical lNeedContext = TypeFactory.logical();
      character cContainerType = TypeFactory.character();
      
      return function("getNeedContext", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("NeedContext"));
         lNeedContext.assign(ghProp.unwrapBufferField().value());
         
         if (_isUnknown(lNeedContext))
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("QueryObject"));
            lNeedContext.assign(ghProp.unwrapBufferField().value());
            
            /* A container that is Appserver Aware need context
                  (Appserver Aware is given by the existence of this super-procedure)*/
            if (_not(lNeedContext))
            {
               silent(() -> 
               {
                  ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                  ghProp.assign(ghProp.unwrapBuffer().bufferField("ContainerType"));
                  cContainerType.assign(ghProp.unwrapBufferField().value());
               });
               
               if (_or(notUnknown(cContainerType), () -> isNotEqual(cContainerType, "")))
               {
                  lNeedContext.assign(true);
               }
            }
            /* {set Prop value [handle]} */ 
            ControlFlowOps.invokeDynamicFunctionInWithMode(logical.class, "setNeedContext", targetProcedure(), "I", lNeedContext);
         }
         returnNormal(lNeedContext);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public character getServerFileName()
   {
      /*------------------------------------------------------------------------------
        Purpose:  returns the property value that stores the actual server-side
                   SDO filename to run on the AppServer -- may not be the ObjectName
                   if that has been modified.
         Params:  <none>
          Notes:  Defaults to the target-procedure file-name without the _cl.      
      ------------------------------------------------------------------------------*/
      character cName = TypeFactory.character();
      logical lQuery = TypeFactory.logical();
      
      return function("getServerFileName", character.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerFileName"));
         cName.assign(ghProp.unwrapBufferField().value());
         
         if (_or(isUnknown(cName), () -> isEqual(cName, "")))
         {
            ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
            ghProp.assign(ghProp.unwrapBuffer().bufferField("QueryObject"));
            lQuery.assign(ghProp.unwrapBufferField().value());
            
            if ((lQuery).booleanValue())
            {
               cName.assign(replaceAll(targetProcedure().unwrapFile().getFileName(), "_cl.", "."));
            }
            else
            {
               cName.assign("adm2/dyncontainer.w");
            }
         }
         returnNormal(cName);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical getServerFirstCall()
   {
      /*------------------------------------------------------------------------------
        Purpose:  Returns (LOGICAL) a flag indicating whether this is the first 
                  call to the server side object.
         Params:  <none>
          Notes:  Server side property! (The client uses AsHasStarted)
                  Typically usage:
                   Client  
                    if not AsHasStarted 
                      contextForServer = ServerFirstCall=YES
                   Server       
                    if ServerFirstCall 
                      contextForClient = AsHasStarted=YES 
      ------------------------------------------------------------------------------*/
      logical lFirstCall = TypeFactory.logical();
      
      return function("getServerFirstCall", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerFirstCall"));
         lFirstCall.assign(ghProp.unwrapBufferField().value());
         returnNormal(lFirstCall);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public character getServerOperatingMode()
   {
      /*------------------------------------------------------------------------------
         Purpose: get the ServerOperatingMode which tells whether the object is 
                  running. 
                    'Stateless',
                    'State-Reset',
                    'State-Aware' or 
                    'none' (running locally)
           Notes:  
       ------------------------------------------------------------------------------*/
      character cServerOperatingMode = TypeFactory.character();
      
      return function("getServerOperatingMode", character.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         cServerOperatingMode.assign(ghProp.unwrapBufferField().value());
         returnNormal(cServerOperatingMode);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical hasNoServerBinding()
   {
      /*------------------------------------------------------------------------------
        Purpose: Encapsulates logic to check AppServer properties to see if the 
                 object has no current or future server bindings and is using a 
                 stateless operating mode.    
          Notes: Used by data objects to check if data can be retrieved with one 
                 appserver hit together with other data objects and by another 
                 object (DataContainer).   
      ------------------------------------------------------------------------------*/
      character cAsdivision = TypeFactory.character();
      logical lAsbound = TypeFactory.logical();
      character cOperatingMode = TypeFactory.character();
      character cBindScope = TypeFactory.character();
      
      return function("hasNoServerBinding", logical.class, new Block((Body) () -> 
      {
         /* {get Prop var [handle]} */ 
         cAsdivision.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getASDivision", targetProcedure()));
         /* {get Prop var [handle]} */ 
         lAsbound.assign(ControlFlowOps.invokeDynamicFunctionIn(logical.class, "getAsBound", targetProcedure()));
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         cOperatingMode.assign(ghProp.unwrapBufferField().value());
         /* {get Prop var [handle]} */ 
         cBindScope.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getBindScope", targetProcedure()));
         returnNormal(and(and(and(isEqual(cAsdivision, "Client"), not(lAsbound)), () -> not(matchesList("state-aware,state-reset", cOperatingMode))), () -> not(matchesList("Strong,this", cBindScope))));
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public handle runServerProcedure(final character _pcServerFileName_1, final handle _phAppService_1)
   {
      character pcServerFileName_1 = TypeFactory.initInput(_pcServerFileName_1);
      handle phAppService_1 = TypeFactory.initInput(_phAppService_1);
      /*------------------------------------------------------------------------------
        Purpose: Returns handle of the (a) procedure after a run on server.  
          Notes: Intended to simplify an override of the run statement for example 
                 to use a bind procedure instead of running the procedure directly. 
      ------------------------------------------------------------------------------*/
      handle hAshandle = TypeFactory.handle();
      
      return function("runServerProcedure", handle.class, new Block((Body) () -> 
      {
         OnPhrase[] onPhrase0 = new OnPhrase[]
         {
            new OnPhrase(Condition.STOP, Action.LEAVE, "blockLabel0")
         };
         
         doBlock(TransactionType.SUB, "blockLabel0", onPhrase0, new Block((Body) () -> 
         {
            ControlFlowOps.invokeRemotePersistentSet((pcServerFileName_1).toStringMessage(), hAshandle, (handle) null, phAppService_1, false);
         }));
         
         returnNormal(hAshandle);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAppService(final character _pcAppService_1)
   {
      character pcAppService_1 = TypeFactory.initInput(_pcAppService_1);
      /*------------------------------------------------------------------------------
        Purpose:    Sets the AppService property which names logical partition name 
                    of the AppService to be used to connect to an AppServer
                     
        Parameters: INPUT pcPartition - 
                       Partition name that this Object has registered.  
        Notes:      The default value is the APPLICATION-PARTITION preprocessor value 
                    which can be defined in the procedure settings dialog in the 
                    AppBuilder.
                    No xp preprocessor in order to fix '(None)' as saved from 
                    the Instance Property dialog or AppBuilder.  
      ------------------------------------------------------------------------------*/
      character cAppService = TypeFactory.character();
      
      return function("setAppService", logical.class, new Block((Body) () -> 
      {
         cAppService.assign(_isEqual(pcAppService_1, "(None)") ? new character("") : pcAppService_1);
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("AppService"));
         ghProp.unwrapBufferField().changeValue(cAppService);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAsdivision(final character _pcDivision_1)
   {
      character pcDivision_1 = TypeFactory.initInput(_pcDivision_1);
      
      return function("setASDivision", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASDivision"));
         ghProp.unwrapBufferField().changeValue(pcDivision_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAshandle(final handle _phAshandle_1)
   {
      handle phAshandle_1 = TypeFactory.initInput(_phAshandle_1);
      
      return function("setASHandle", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHandle"));
         ghProp.unwrapBufferField().changeValue(phAshandle_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAsHasStarted(final logical _plHasStarted_1)
   {
      logical plHasStarted_1 = TypeFactory.initInput(_plHasStarted_1);
      
      return function("setAsHasStarted", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASHasStarted"));
         ghProp.unwrapBufferField().changeValue(plHasStarted_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAsinfo(final character _pcInfo_1)
   {
      character pcInfo_1 = TypeFactory.initInput(_pcInfo_1);
      
      return function("setASInfo", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASInfo"));
         ghProp.unwrapBufferField().changeValue(pcInfo_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAsinitializeOnRun(final logical _plInitialize_1)
   {
      logical plInitialize_1 = TypeFactory.initInput(_plInitialize_1);
      
      return function("setASInitializeOnRun", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASInitializeOnRun"));
         ghProp.unwrapBufferField().changeValue(plInitialize_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setAsusePrompt(final logical _plFlag_1)
   {
      logical plFlag_1 = TypeFactory.initInput(_plFlag_1);
      
      return function("setASUsePrompt", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ASUsePrompt"));
         ghProp.unwrapBufferField().changeValue(plFlag_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setBindScope(final character _pcScope_1)
   {
      character pcScope_1 = TypeFactory.initInput(_pcScope_1);
      /*------------------------------------------------------------------------------
        Purpose: Define the scope of an stateless appserver connection 
      Parameter: pcScope - A character string defining the scope 
                     The order below also indicates the scope 'strength' with the 
                     weakest scopes first, this is important as a weaker scope never 
                     can override a stronger or even itself. (blank is the exception)    
                     - Blank (Default) 
                       Unbind at the level where the bind takes place.                 
                       It overrides all other scopes and should NOT need to be used 
                       except by unbindServer itself in order to reset to default!     
                       - getAsHandle and bindServer will set BindSignature which 
                         then is used by unbindServer.
                       - A 'one-hit' DataRequest will be used if not already bound.   
                     - 'Data' 
                       Unbind immediately after a data request, but keep any 
                       connection that happens before the data request. 
                        - This is enforced by endClientDataReqest or implicit by 
                          allowing a 'one-hit' datarequest. 
                        - unbindServer will unbind at the first/upper level it was 
                          set if not dataRequest took place.    
                        - A 'one hit' datarequest can be used if not bound.      
                     - 'This'  
                       Don't unbind until we are back at this level. 
                         This is different from bindServer as no connection is being 
                         done until required.
                        - Prevents a 'one hit' datarequest 
                     - 'Strong'  
                       Only unbind when unbindServer('unconditional')                 
                     - ? 
                       This is the weakest scope, but the strongest rule and should 
                       really only be used to reset the value to blank. 
                                      
          Notes: The property is enforced by endClientdataRequest and unbindServer
                 or in some cases in an actual dataRequest. 
               - unbindServer should ALWAYS be called at the end of a procedure where 
                 this is set, to ensure that an unbind takes place under all 
                 circumstances and that the scope is reset to blank. 
                 The only exception being the STRONG connection, which relies on an 
                 unbindServer('unconditional') further up in the call stack.        
               - endClientdataRequest is normally used further down and will unbind
                 if the scope is 'Data'. (It also ensures that context is retrieved 
                 from server if it is still unbound)            
               - The BindSignature property is used to identify the scope level and
                 is set here unless blank, in which case it need to be set in 
                 bindServer and getAsHandle for unbindServer to enforce the rule.   
      ------------------------------------------------------------------------------*/
      character cCurrentScope = TypeFactory.character();
      
      return function("setBindScope", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("BindScope"));
         cCurrentScope.assign(ghProp.unwrapBufferField().value());
         
         switch (StringHelper.safeTrimTrailing((pcScope_1).toStringMessage()).toUpperCase())
         {
            case "":
               {
                  ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                  ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
                  ghProp.unwrapBufferField().changeValue(new unknown());
                  ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                  ghProp.assign(ghProp.unwrapBuffer().bufferField("BindScope"));
                  ghProp.unwrapBufferField().changeValue(new character(""));
               }
               break;
            case "STRONG":
               {
                  ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                  ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
                  ghProp.unwrapBufferField().changeValue(new unknown());
                  ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                  ghProp.assign(ghProp.unwrapBuffer().bufferField("BindScope"));
                  ghProp.unwrapBufferField().changeValue(pcScope_1);
               }
               break;
            case "THIS":
               {
                  if (_not(matchesList("Strong,This", cCurrentScope)))
                  {
                     ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                     ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
                     ghProp.unwrapBufferField().changeValue(EnvironmentOps.getSourceName(2));
                     ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                     ghProp.assign(ghProp.unwrapBuffer().bufferField("BindScope"));
                     ghProp.unwrapBufferField().changeValue(pcScope_1);
                  }
               }
               break;
            case "DATA":
               {
                  if (_not(matchesList("Strong,This,Data", cCurrentScope)))
                  {
                     ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                     ghProp.assign(ghProp.unwrapBuffer().bufferField("BindSignature"));
                     ghProp.unwrapBufferField().changeValue(EnvironmentOps.getSourceName(2));
                     ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
                     ghProp.assign(ghProp.unwrapBuffer().bufferField("BindScope"));
                     ghProp.unwrapBufferField().changeValue(pcScope_1);
                  }
               }
               break;
            default:
               {
                  returnNormal(false);
               }
         }
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setNeedContext(final logical _plNeedContext_1)
   {
      logical plNeedContext_1 = TypeFactory.initInput(_plNeedContext_1);
      
      return function("setNeedContext", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("NeedContext"));
         ghProp.unwrapBufferField().changeValue(plNeedContext_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setServerFileName(final character _pcFileName_1)
   {
      character pcFileName_1 = TypeFactory.initInput(_pcFileName_1);
      
      return function("setServerFileName", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerFileName"));
         ghProp.unwrapBufferField().changeValue(pcFileName_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setServerFirstCall(final logical _plFirstCall_1)
   {
      logical plFirstCall_1 = TypeFactory.initInput(_plFirstCall_1);
      
      return function("setServerFirstCall", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerFirstCall"));
         ghProp.unwrapBufferField().changeValue(plFirstCall_1);
         returnNormal(true);
      }));
   }

   /**
    * _UIB-CODE-BLOCK-END
    */
   public logical setServerOperatingMode(final character _pcServerOperatingMode_1)
   {
      character pcServerOperatingMode_1 = TypeFactory.initInput(_pcServerOperatingMode_1);
      /*------------------------------------------------------------------------------
         Purpose: set the serverOperatingMode.   
      Parameters: pcServerOperatingMode can be:
                    'Stateless'
                    'State-Reset'
                    'State-Aware' or
                    'none'        (when running locally)
           Notes:  The xp Preprocessor IS defined, so the protection against 
                   'none' only works for outside callers! 
                   (WebSpeed specifically ) 
                   The instance property can be forced to STATE-RESET.
      ------------------------------------------------------------------------------*/
      character cCurrentMode = TypeFactory.character();
      character cAsDivision = TypeFactory.character();
      
      return function("setServerOperatingMode", logical.class, new Block((Body) () -> 
      {
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         cCurrentMode.assign(ghProp.unwrapBufferField().value());
         
         /* 'Stateless' is set from the AppServer, it can be forced to 'state-reset',
              but must NOT be set to NONE when this is a 'client' 
              See notes above. */
         if (_and(and(begins(cCurrentMode, "STATE"), () -> isEqual(pcServerOperatingMode_1, "NONE")), () -> isNotEqual(sourceProcedure(), targetProcedure())))
         {
            /* {get Prop var [handle]} */ 
            cAsDivision.assign(ControlFlowOps.invokeDynamicFunctionIn(character.class, "getASDivision", targetProcedure()));
            
            if (_isEqual(cAsDivision, "Client"))
            {
               returnNormal(false);
            }
         }
         ghProp.assign(handle.fromString(entry(1, targetProcedure().unwrapADMData().getADMData(), chr(1))));
         ghProp.assign(ghProp.unwrapBuffer().bufferField("ServerOperatingMode"));
         ghProp.unwrapBufferField().changeValue(pcServerOperatingMode_1);
         returnNormal(true);
      }));
   }

   public character fwdGetAdmversion()
   {
      return function("fwdGetADMVersion", character.class, new Block((Body) () -> 
      {
         returnNormal("ADM2.2");
      }));
   }
}
