/**
 * API for the aster plot control.
 *
 * Renders an SVG aster plot from an array of JSON objects using D3. The expected data members of
 * each object are (at minimum):
 *
 * {number} id     -  Unique identifier for object.
 * {number} weight -  Value between 0 and 100. Determines relative length of an arc.
 * {number} score  -  Value between 0 and 100. Determines relative, outer radius of an arc.
 * {string} color  -  Color specifier; fill color of an arc.
 * {string} label  -  Label for an arc of the plot; used as a tool tip.
 *
 * The weighted average of all scores is displayed as a percentage in the center of the plot.
 *
 * Derived from a sample by Ben Best (http://bl.ocks.org/bbest/2de0e25d4840c68f2db1), which had
 * contributions from Jim Regetz (https://github.com/regetz/aster-plot) and Parker Abercrombie
 * (https://github.com/parkerabercrombie), and based on work by Mike Bostock
 * (http://bl.ocks.org/mbostock/3887193).
 *
 * This work contains significant changes by Golden Code Development.
 *
 * Released under the MIT License (https://opensource.org/licenses/MIT).
 */
define(function()
{
   /** Object which contains API functions. */
   var me = {};
   
   /**
    * Generate the aster plot under the given DOM element, using the given data.
    *
    * @param   {DOM element} parent
    *          DOM element to which to attach the SVG
    */
   me.generate = function(parent, data)
   {
      var width = 275,
          height = 275,
          radius = Math.min(width, height) / 2,
          innerRadius = 0.3 * radius;
      
      // sort arcs by weight
      var pie = d3.pie()
                  .sort(null)
                  .value(function(d) { return d.weight; });
      
      // set up tool tips
      var tip = d3.tip()
                  .attr('class', 'd3-tip')
                  .offset([0, 0])
                  .html(function(d)
                  {
                     return d.data.label + ": <span style='color:orangered'>" + d.data.score + "</span>";
                  });
      
      // inner arcs
      var arc = d3.arc()
                  .innerRadius(innerRadius)
                  .outerRadius(function (d)
                  {
                     return (radius - innerRadius) * (d.data.score / 100.0) + innerRadius; 
                  });
      
      // outer arcs
      var outlineArc = d3.arc()
                         .innerRadius(innerRadius)
                         .outerRadius(radius);
      
      // main svg
      var svg = d3.select(parent).append("svg")
                  .attr("width", width)
                  .attr("height", height)
                  .append("g")
                  .attr("transform", "translate(" + width / 2 + "," + height / 2 + ")");
      
      svg.call(tip);
      
      // prepare data
      data.forEach(function(d)
      {
         d.id     =  d.id;
         d.color  =  d.color;
         d.weight = +d.weight;
         d.score  = +d.score;
         d.label  =  d.label;
      });
      
      // apply data to arcs
      var path = svg.selectAll(".solidArc")
                    .data(pie(data))
                    .enter().append("path")
                    .attr("fill", function(d) { return d.data.color; })
                    .attr("class", "solidArc")
                    .attr("d", arc)
                    .on('mouseover', tip.show)
                    .on('mouseout', tip.hide);
      var outerPath = svg.selectAll(".outlineArc")
                         .data(pie(data))
                         .enter().append("path")
                         .attr("fill", "none")
                         .attr("class", "outlineArc")
                         .attr("d", outlineArc);
      
      // calculate the weighted mean score
      var score = 
         data.reduce(function(a, b)
         {
            return a + (b.score * b.weight); 
         }, 0) / 
         data.reduce(function(a, b)
         {
            return a + b.weight; 
         }, 0);
      
      // central text
      svg.append("svg:text")
         .attr("class", "aster-score")
         .attr("dy", ".35em")
         .attr("text-anchor", "middle")
         .text("" + Math.round(score) + "%");
   }
   
   return me;
});
