/*
 * Copyright 2016-2017 Golden Code Development Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"),  to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * This function is loaded by the Hotel embedded web application's index.html. It defines the
 * primary JavaScript API for the application.
 */
define(function ()
{
   "use strict";
   
   /** Object which contains API functions. */
   var me = {};
   
   /**
    * The name of the program where the APIs are located.  This will be ran persistent when the
    * window is selected to be shown.
    */
   var remoteProgram = "ehotel.p";
   
   /** The handle for the remote program, in which the APIs are executed. */
   var remoteProgramHandle;

   /** Current start date for room utilization by type data. */
   var utilByTypeStartDate;
   
   /** Current end date for room utilization by type data. */
   var utilByTypeEndDate;
   
   /** Resource ID for a 'roomUtilization' pub/sub event published by converted ABL code. */
   var roomUtilizationResourceId;
   
   /** Resource ID for a 'capacityZoom' pub/sub event published by JavaScript code. */
   var capacityZoomResourceId;
   
   /** Resource ID for a 'roomTypePick' pub/sub event published by JavaScript code. */
   var roomTypePickResourceId;
   
   /** The calendar object. */
   var calendar;
   
   /** The aster object. */
   var aster;
      
   /**
    * Format the given date into the format expected by ABL (or return today if no date is passed).
    *
    * @param   {Date} date
    *          The date to format.
    */
   function formattedDate(date)
   {
      if (!date)
      {
         date = new Date(Date.now());
      }
      
      return moment(date).format("MM/DD/YYYY");
   };
   
   /**
    * Parse a Progress date string (in the format 99/99/999) and return the corresponding date.
    * 
    * TODO: make this honor the current SESSION:DATE-FORMAT setting (mdy/dmy); currently, it
    * assumes mdy.
    */
   function parseProgressDate(dateString)
   {
      // assume Progress date format of 99/99/9999 (mdy)
      var a = dateString.split(/[^0-9]/);
      
      return new Date(a[2], a[0] - 1, a[1]);
   };
   
   /**
    * Callback executed when the P2J page has loaded.  It runs the remote program persistent.
    * @param    {function} initCallback
    *           Callback executed once initialization is done.
    * @param    {object} vcalendar
    *           The calendar object.
    * @param    {object} vaster
    *           The aster object.
    */
   me.initialize = function(initCallback, vcalendar, vaster)
   {
      calendar = vcalendar;
      aster = vaster;
      
      p2j.embedded.registerIncomingRequest("utilizationByDateResponse", utilizationByDateResponse, parseUtilizationByDateResponse);
      p2j.embedded.registerIncomingRequest("utilizationByTypeResponse", utilizationByTypeResponse, parseUtilizationByTypeResponse);

      // do not call the remote program here, wait for the "p2jInitialized" notification
      p2j.embedded.silentErrorEnable();

      // examples for UNSUBSCRIBE and PUBLISH:
      // p2j.embedded.unsubscribe(unsubscribeCallback, true, loggedinResourceEvent, "loggedin", undefined);
      // p2j.embedded.publish("loggedin", loggedinResourceEvent, ["hotel2", "hotel2"]);
      
      // subscribe to a "roomUtilization" event published by converted ABL code
      p2j.embedded.subscribe(subscribeRoomUtilizationCallback,
                             true,
                             undefined, // this was not created yet, will be saved during callback
                             "roomUtilization",
                             undefined,
                             roomUtilizationEvent);
      
      // create a resource for a "capacityZoom" event that will be published from JavaScript
      p2j.embedded.createResource(capacityZoomResourceCreated);
      
      // create a resource for a "roomTypePick" event that will be published from JavaScript
      p2j.embedded.createResource(roomTypePickResourceCreated);
      
      p2j.embedded.invokeExternalProgramPersistent(function(msg)
      {
         if (msg.payload.error)
         {
            dbg(msg.payload["error-num"] + " " + msg.payload["error-msg"]);
         }

         remoteProgramHandle = msg.payload.handle;
         
         activateWindow();
         
         if (initCallback)
         {
            initCallback();
         }
      },
      remoteProgram, {});
   };
   
   /**
    * Get the DOM element used to activate the specified page.
    *
    * @param    {number} pageId
    *           The 1-indexed page number.
    * @param    {string} pageTitle
    *           The page title as collected from the legacy code.
    */
   me.getPageButton = function(pageId, pageTitle)
   {
      switch (pageId)
      {
         case 1:
            return $("<li><a id=\"fwd-page-1\" href=\"#fwd-embedded-page\" style=\"display: none\"><span class=\"glyphicon glyphicon-list-alt\"></span> Available</a></li>");
         case 2:
            return $("<li><a id=\"fwd-page-2\" href=\"#fwd-embedded-page\" style=\"display: none\"><span class=\"glyphicon glyphicon-user\"></span> Guests</a></li>");
         case 3:
            return $("<li><a id=\"fwd-page-3\" href=\"#fwd-embedded-page\" style=\"display: none\"><span class=\"glyphicon glyphicon-calendar\"></span> Reservations</a></li>");
         case 4:
            return $("<li><a id=\"fwd-page-4\" href=\"#fwd-embedded-page\" style=\"display: none\"><span class=\"glyphicon glyphicon-scale\"></span> Rates</a></li>");
         case 5:
            return $("<li><a id=\"fwd-page-5\" href=\"#fwd-embedded-page\" style=\"display: none\"><span class=\"glyphicon glyphicon-bed\"></span> Rooms</a></li>");
      }
   };

   me.windowDestroyed = function()
   {
      p2j.embedded.silentErrorEnable();

      p2j.embedded.unsubscribe(undefined, true, roomUtilizationResourceId, "roomUtilization", undefined);
      
      p2j.embedded.deleteHandle(roomUtilizationResourceId);
      p2j.embedded.deleteHandle(roomTypePickResourceId);
      p2j.embedded.deleteHandle(capacityZoomResourceId);

      roomUtilizationResourceId = undefined;
      roomTypePickResourceId = undefined;
      capacityZoomResourceId = undefined;
      
      p2j.embedded.deleteHandle(remoteProgramHandle);
      remoteProgramHandle = undefined;

      p2j.embedded.silentErrorDisable();
      
      $("#ehotelAster").css("display", "none");
      $("#ehotelCalendar").css("display", "none");

      p2j.embedded.deregisterIncomingRequest("utilizationByDateResponse");
      p2j.embedded.deregisterIncomingRequest("utilizationByTypeResponse");
   };
   
   me.windowCreated = function()
   {
      $("#ehotelAster").css("display", "");
      $("#ehotelCalendar").css("display", "");
   };

   /**
    * Activate this window by populating the additional widgets.
    */
   function activateWindow()
   {
      // get the initial set of room by type data for the current date, to populate the aster
      // plot and tabulator table
      var d = new Date();
      getUtilizationByType(d, d);

      // initialize the tabulator
      $("#utilTypeTable").tabulator(
      {
         fitColumns: true, // fit columns to width of table
         height: "100%",
         columns:
         [
            {
               title:"Type",
               field:"label",
               tooltip:true,
               tooltipHeader:"Room Type",
               sorter:"string",
            },
            {
               title:"Room Nights",
               columns:
               [
                  {
                     title:"Actual",
                     field:"roomsOccupied",
                     tooltipHeader:"Rented Room Nights",
                     sorter:"number",
                     align:"right",
                     width:75,
                  },
                  {
                     title:"Total",
                     field:"totalRooms",
                     tooltipHeader:"Total Room Nights",
                     sorter:"number",
                     align:"right",
                     width:75,
                  },
               ],
            },
            {
               title:"Revenue",
               columns:
               [
                  {
                     title:"Actual",
                     field:"actualRevenue",
                     tooltip:true,
                     tooltipHeader:"Actual Revenue",
                     sorter:"number",
                     align:"right",
                     width:90,
                  },
                  {
                     title:"Poten.",
                     field:"potentialRevenue",
                     tooltip:true,
                     tooltipHeader:"Potential Total Revenue",
                     sorter:"number",
                     align:"right",
                     width:90,
                  },
               ],
            },
            {
               title:"%",
               field:"score",
               tooltip:true,
               tooltipHeader:"Room Type Utilization",
               sorter:"number",
               formatter:"progress",
               width:50,
            },
         ],
         rowClick:function(e, id, data, row)
         {
            // notify subscribers that a room type has been picked in table
            var i = id.toString();
            var a = formattedDate(utilByTypeStartDate);
            var b = formattedDate(utilByTypeEndDate);
            p2j.embedded.publish("roomTypePick", roomTypePickResourceId, [i, a, b]);
         },
      });
      
      // re-render table on window resize
      $(window).resize(function()
      {
         $("#utilTypeTable").tabulator("redraw", true);
      });

      // get date range calendar can render within its parent's width
      var calDates = calendar.initialize(new Date(), $("#capacityCalParent").width());
      
      // query daily utilization data
      getUtilizationByDate(formattedDate(calDates[0]), formattedDate(calDates[1]));
   };
   
   /**
    * Callback executed for an UNSUBSCRIBE response.
    *
    * Not currently in use; provided as an example.
    *
    * @param   {object} msg
    *          The response message.
    */
   function unsubscribeCallback(msg)
   {
      var payload = msg["payload"];
      if (payload["error"])
      {
         console.log(payload["error-num"] + " " + payload["error-msg"]);
      }
      p2j.emain.showMessage("unsubscribeCallback", msg);
   };
   
   /**
    * Callback executed for a 'roomUtilization' subscribe response.
    *
    * @param   {object} msg
    *          The response message.
    */
   function subscribeRoomUtilizationCallback(msg)
   {
      var payload = msg["payload"];
      if (payload["error"])
      {
         console.log(payload["error-num"] + " " + payload["error-msg"]);
      }

      p2j.emain.showMessage("subscribeRoomUtilizationCallback", msg);

      if (roomUtilizationResourceId)
      {
         return;
      }
      
      roomUtilizationResourceId = msg["payload"]["subscriber-id"];
   };
   
   /**
    * Callback executed for creation of a resource associated with the 'capacityZoom' pub/sub
    * event.
    *
    * @param   {object} msg
    *          The response message.
    */
   function capacityZoomResourceCreated(msg)
   {
      p2j.emain.showMessage("capacityZoomResourceCreated", msg);
      var payload = msg["payload"];
      capacityZoomResourceId = msg["payload"]["resource-id"];
   };
   
   /**
    * Callback executed for creation of a resource associated with the 'roomTypePick' pub/sub
    * event.
    *
    * @param   {object} msg
    *          The response message.
    */
   function roomTypePickResourceCreated(msg)
   {
      p2j.emain.showMessage("roomTypePickResourceCreated", msg);
      var payload = msg["payload"];
      roomTypePickResourceId = msg["payload"]["resource-id"];
   };
   
   /**
    * Callback executed for the "roomUtilization" PUBLISH event. This event is published by
    * converted ABL code when either a new hotel check-in or a new hotel reservation is booked.
    *
    * @param   {array} params
    *          The publish event parameters.
    */
   function roomUtilizationEvent(params)
   {
      // staywrtrg.p can be triggered on cancel of Check-in dialog (with an empty record)
      if (params[0] == "?" || params[1] == "?")
      {
         return;
      }
      // start date is check-in date of stay/reservation
      var start = parseProgressDate(params[0]);
      
      // end date is day before check-out date of stay/reservation
      var end = moment(parseProgressDate(params[1])).subtract(1, "days").toDate();
      
      // up-call to collect utilization/capacity data for given date window
      getUtilizationByDate(formattedDate(start), formattedDate(end));
      
      // up-call to collect utilization data by room type for given date window; only invoke
      // if the currently displayed date window overlaps with the period that changed
      if (utilByTypeStartDate <= end && utilByTypeEndDate >= start)
      {
         getUtilizationByType(utilByTypeStartDate, utilByTypeEndDate);
      }
   };
   
   /**
    * Send a command to load the room availability table.
    *
    * @param   {String} adate
    *          The date on which the availability is computed.
    */
   function loadRooms(adate)
   {
      var args =
      {
         "adate" : adate
      }
      
      p2j.emain.invoke(remoteProgramHandle, "loadRooms", args);
   };
   
   /**
    * Make an up-call to converted ABL code to collect room utilization data, grouped by room
    * type, aggregated for the specified date range.
    *
    * @param   {date} start
    *          Starting date.
    * @param   {date} end
    *          Ending date.
    */
   function getUtilizationByType(start, end)
   {
      var args = 
      {
         "start-date" : formattedDate(start),
         "end-date"   : formattedDate(end)
      };

      p2j.emain.invoke(remoteProgramHandle, "utilizationByType", args);
   };
   
   /**
    * Make an up-call to converted ABL code to collect room utilization data for each date in
    * the specified range.
    *
    * @param   {string} start
    *          A Progress-formatted date string for the start of the range.
    * @param   {string} end
    *          A Progress-formatted date string for the end of the range.
    */
   function getUtilizationByDate(start, end)
   {
      var args = 
      {
         "start-date" : start,
         "end-date"   : end
      };

      p2j.emain.invoke(remoteProgramHandle, "utilizationByDate", args);
   };
   
   /**
    * This function is invoked from within the calendar heat map when the user clicks a date or
    * swipes a date range in the calendar control. It updates the aster plot, room utilization
    * table, and available rooms embedded content view (if showing) to display information
    * specific to the selected date range.
    *
    * This causes a 'capacityZoom' event to be published.
    *
    * @param   {date} start
    *          Starting date for the selected range; will always be less than or equal to end
    *          date, regardless of the direction the user swiped.
    * @param   {date} end
    *          Ending date for the selected range; will always be greater than or equal to start
    *          date, regardless of the direction the user swiped.
    *          
    */
   function dateRangeSelected(start, end)
   {
      // refresh aster plot and room utilization table of room utilization by type
      getUtilizationByType(start, end);
      
      // notify subscribers that a date range on the hotel capacity heatmap has been selected
      var a = formattedDate(start);
      var b = formattedDate(end);
      p2j.embedded.publish("capacityZoom", capacityZoomResourceId, [a, b]);
   };
      
   /**
    * Parser for the "sendMsg_utilizationByDateResponse" REMOTE_CALL request.
    *
    * @param   {string} msg
    *          The incoming message.
    *
    * @return  {object}
    *          The JSON object.
    */
   function parseUtilizationByDateResponse(msg)
   {
      var jsonMsg = JSON.parse(msg);
      var messageId = jsonMsg.messageId;
      var omsg =
      {
         "messageId"  : parseInt(messageId),
         "rows"       : jsonMsg.rows,
         "domain"     : jsonMsg.domain,
         "startDate"  : jsonMsg.start,
         "endDate"    : jsonMsg.end
      };
      
      return omsg;
   };
   
   /**
    * Parser for the "sendMsg_utilizationByTypeResponse" REMOTE_CALL request.
    *
    * @param   {string} msg
    *          The incoming message.
    *
    * @return  {object}
    *          The JSON object.
    */
   function parseUtilizationByTypeResponse(msg)
   {
      var jsonMsg = JSON.parse(msg);
      var messageId = jsonMsg.messageId;
      var omsg =
      {
         "messageId"    : parseInt(messageId),
         "rows"         : jsonMsg.rows,
         "rateWeighted" : (jsonMsg.rateWeighted == "true"),
         "start"        : jsonMsg.start,
         "end"          : jsonMsg.end
      };
      
      return omsg;
   };
   
   /**
    * Handler for the "sendMsg_utilizationByDateResponse" REMOTE_CALL request.
    *
    * Updates "heat map" calendar with utilization data for each day displayed by calendar.
    *
    * @param   {object} msg
    *          The JSON from the parsed message.
    */
   function utilizationByDateResponse(jsonMsg)
   {
      // this will create the calendar control if needed, or update data if already present
      calendar.render(jsonMsg,
                      document.getElementById("capacityCalendar"),
                      dateRangeSelected,
                      "EMPTY",
                      "FULL");
   };

   /**
    * Handler for the "sendMsg_utilizationByTypeResponse" REMOTE_CALL request.
    *
    * Draws an aster plot of utilization by room type. Populates utilization table with data.
    *
    * @param   {object} msg
    *          The JSON from the parsed message.
    */
   function utilizationByTypeResponse(jsonMsg)
   {
      // remove any previously generated plot
      $("#asterPlot").empty();
      
      utilByTypeStartDate = parseProgressDate(jsonMsg.start);
      utilByTypeEndDate = parseProgressDate(jsonMsg.end);
      
      var fmt = "ddd MMM DD YYYY";
      var txt = moment(utilByTypeStartDate).format(fmt);
      if (utilByTypeStartDate < utilByTypeEndDate)
      {
         txt = txt + " - " + moment(utilByTypeEndDate).format(fmt);
      }
      $("#utilTypePeriod").text(txt);
      
      var rows = jsonMsg.rows;
      
      /* TODO: kludgy, will fail if > 14 room types */
      var colors =
      [
         "#9E0041",
         "#C32F4B",
         "#E1514B",
         "#F47245",
         "#FB9F59",
         "#FEC574",
         "#FAE38C",
         "#EAF195",
         "#C7E89E",
         "#9CD6A4",
         "#6CC4A4",
         "#4D9DB4",
         "#4776B4",
         "#5E4EA1"
      ];
      
      // assign arc colors
      for (var i = 0; i < rows.length; i++)
      {
         rows[i].color = colors[i];
      }
      
      // determine which title to use
      if (jsonMsg.rateWeighted == true)
      {
         $("#utilRoomTypeLabel").text("MEDIAN REVENUE EARNED BY ROOM TYPE");
      }
      else
      {
         $("#utilRoomTypeLabel").text("MEDIAN OCCUPANCY BY ROOM TYPE");
      }
      
      // generate the new plot
      aster.generate("#asterPlot", rows);
      
      // set data into utilization table
      $("#utilTypeTable").tabulator("setData", rows);
      $("#utilTypeTable").tabulator("redraw", true);
   };
   
   return me;
});
