/*
 * Copyright 2016-2018 Golden Code Development Corporation
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy of this software
 * and associated documentation files (the "Software"),  to deal in the Software without
 * restriction, including without limitation the rights to use, copy, modify, merge, publish,
 * distribute, sublicense, and/or sell copies of the Software, and to permit persons to whom the
 * Software is furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all copies or
 * substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
 * BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM,
 * DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
 */

/**
 * This function is loaded by the Hotel embedded web application's index.html. It defines the
 * primary JavaScript API for the application.
 */
define(function ()
{
   "use strict";
   
   /** Object which contains API functions. */
   var me = {};
   
   /**
    * Load the custom JS script (if any) for the specified window and show it.
    *
    * @param    {string}  windowId
    *           The window ID.
    * 
    * @return   <code>true</code> if the window ID is valid.
    */
   me.loadWindowScript = function(windowId)
   {
      p2j.emain.showLoading();

      switch (windowId)
      {
         case "main-window.w":
            require(["custom/ehotel", "custom/calendar", "custom/aster", "dojo/dom", "dojo/dom-construct", "dojo/on", "dojo/domReady!"],
                    function(oeh, cal, aster, dom, domConstruct, on)
                    {
                       p2j.emain.registerActiveWindow(windowId, oeh);
                       
                       if (oeh.initialize)
                       {
                          oeh.initialize(function(){ p2j.embedded.showWindow(windowId, p2j.emain.windowCreated); }, 
                                         cal, aster);
                       }
                       else
                       {
                          p2j.embedded.showWindow(windowId, p2j.emain.windowCreated);
                       }
                    });
            return true;
      }
      
      if (p2j.embedded.isEmbeddedWindow(windowId))
      {
         p2j.embedded.showWindow(windowId, p2j.emain.windowCreated);
         return true;
      }
      
      return false;
   };

   /**
    * For the "p2jInitialized" REMOTE_CALL notification - initializes the ADM windows.
    *
    * @param   {object} msg
    *          An empty object.
    */
   me.p2jInitialized = function(msg)
   {
      var cfg = 
      {
         "menubar" :
         {
            "container" : "fwdMenu"
         },
         "initCallback"    : function() { p2j.emain.showCurrentWindow(p2j.embedded.getCurrentWindow()); },
         "populateWindows" : populateWindows,
         "populatePages"   : populatePages,
         "windowDestroyed" : p2j.emain.windowDestroyed,

         "embedded-status-area"    : true,
         "show-status-callback"    : p2j.emain.showStatus,

         "embedded-message-area"   : true,
         "show-messages-callback"   : p2j.emain.showMessages
      };

      p2j.embedded.initializeEmbeddedWindows(cfg);
   };

   /*
    * Populate the window selection combo.
    *
    * @param    {array} windowDetails
    *           An array with the window details, as retrieved from <code>adm_windows.json</code>.
    */
   function populateWindows(windowDetails)
   {
      windowDetails.sort(function(a, b)
      {
         var t1 = a["window-title"].trim().toLowerCase();
         var t2 = b["window-title"].trim().toLowerCase();

         return t1.localeCompare(t2);
      });

      var sel = $("#embeddedWindows");
      
      for (var i = 0; i < windowDetails.length; i++)
      {
         var title = windowDetails[i]["window-title"];
         var program = windowDetails[i]["program-name"];
         var opt = $('<option></option');
         opt.val(program).html(title);
         
         sel.append(opt);
      }
      
      sel.change(function()
      {
         var val = $(this).find('option:selected').val();
         
         p2j.embedded.showCurrentWindow(wid);
      });
      
      if (windowDetails.length > 1)
      {
         // show only if more than one window.
         sel.css("display", "");
      }
      else
      {
         p2j.embedded.setCurrentWindow(windowDetails[0]["program-name"]);
      }
   };
   
   /**
     * Create the pages for this window.  If the window has custom pages, it will add them last,
     * before the logout button.
     *
     * @param    {string} windowId
     *           The window ID.
     * @param    {array} pages 
     *           The page names.
     */
   function populatePages(windowId, pages)
   {
      $('#topNavMenu').collapse("hide");

      var pageContainer = $("#pageContainer");
      pageContainer.empty();
      
      var wjs = p2j.emain.getWindowScript();
      
      for (var pnum = 0; pnum < pages.length; pnum++)
      {
         var pageTitle = pages[pnum];
         
         var pageButton;
         
         if (wjs && wjs.getPageButton)
         {
            pageButton = wjs.getPageButton(pnum + 1, pageTitle);
         }
         else
         {
            pageButton = $("<li><a id=\"fwd-page-" + (pnum + 1) + "\" href=\"#fwd-embedded-page\" style=\"display: none\">" + pageTitle + " </a></li>");
         }
         
         pageContainer.append(pageButton);
      }
      
      // additional pages.
      if (wjs && wjs.populateCustomPages)
      {
         wjs.populateCustomPages(pageContainer);
      }
      
      var logoutButton = getLogoutButton();
      pageContainer.append(logoutButton);
      
      // show navbar menu items
      $(".nav > li a", "#topNavBar").show();

      $(".nav > li > a", "#topNavMenu").on("click", function(evt) 
      {
         $.proxy(showEmbeddedPage, this, evt)();
      });
   };
   
   /**
    * Get the logout button DOM element.
    *
    * @return   See above.
    */
   function getLogoutButton()
   {
      return $("<li><a id=\"fwd-page-logout\" href=\"#fwd-embedded-page\" style=\"display: none\"><span class=\"glyphicon glyphicon-log-out\"></span> Logout</a></li>");
   };

   /**
     * Show the embedded page targeted by this event. 
     *
     * @param    {Event} evt
     *           The event object.
     */
   function showEmbeddedPage(evt)
   {
      var id = this.id;
      var pageId;
      
      if (id.startsWith("fwd-page-"))
      {
         pageId = id.substring("fwd-page-".length);
      }
      else
      {
         return;
      }
      
      evt.preventDefault();
      $(".nav > li.active", "#topNavBar").toggleClass("active");
      $("#" + this.id).parent().toggleClass("active");

      var wsj = p2j.emain.getWindowScript();
      
      if (pageId == "logout")
      {
         p2j.embedded.logout(p2j.emain.refreshPage);
      }
      else if (wsj && wsj.showEmbeddedPage && wsj.showEmbeddedPage(pageId))
      {
         // handled...
      }
      else
      {
         p2j.emain.showLoading();
         p2j.embedded.showWindowPage(pageId, p2j.emain.hideLoading);
      }

      $('#topNavMenu').collapse("hide");
   };
   

   /**
    * Show the 'Loading...' overlay and hide the iframe.
    */
   function showLoading()
   {
   };
   me.showLoading = showLoading;

      
   /**
    * Hide the 'Loading...' overlay and show the iframe.
    */
   function hideLoading()
   {
      $("#bgFader").show();
      $("#waitMsg").css("opacity", 0);
   };
   me.hideLoading = hideLoading;

   return me;
});
